from collections.abc import Callable
from math import cos, pi, radians, sqrt

from ...array import argpass
from ...types import Point2D
from ..cartesian import (
    cartesian,
    cartesian_add_in_place,
    cartesian_cross,
    cartesian_dot,
    cartesian_scale,
    spherical,
)
from ..circle import circle_stream
from ..common import LineStream, PolygonStream, Stream
from ..point_equal import point_equal
from .clip import Clip, clip

EPSILON = 1e-6


class ClipCircle(LineStream):
    def __init__(
        self,
        radius: float,
        cr: float,
        delta: float,
        small_radius: bool,
        not_hemisphere: bool,
        stream: Stream,
    ):
        self._radius = radius
        self._cr = cr
        self._delta = delta
        self._small_radius = small_radius
        self._not_hemisphere = not_hemisphere

        self._stream = stream
        self._point0 = None
        self._c0 = None
        self._v0 = None
        self._v00 = None
        self._clean = None

    def line_start(self):
        self._v00 = False
        self._v0 = False
        self._clean = 1

    def point(self, lambda_: float, phi: float):
        point1 = [lambda_, phi]
        v = cos(lambda_) * cos(phi) > self._cr
        c = 0
        if self._small_radius and not v:
            c = self._code(lambda_, phi)
        elif v:
            cst = pi if lambda_ < 0 else -pi
            c = self._code(lambda_ + cst, phi)

        if not self._point0:
            self._v00 = v
            self._v0 = v
            if v:
                self._stream.line_start()
        if v != self._v0:
            point2 = self._intersect(self._point0, point1)
            if (
                not point2
                or point_equal(self._point0, point2)
                or point_equal(point1, point2)
            ):
                if len(point1) == 2:
                    point2.append(1)
                else:
                    point1[2] = 1

        if v != self._v0:
            self._clean = 0
            if v:
                self._stream.line_start()
                point2 = self._intersect(point1, self._point0)
                self._stream.point(point2[0], point2[1])
            else:
                point2 = self._intersect(self._point0, point1)
                argpass(self._stream.point)(point2[0], point2[1], 2)
                self._stream.line_end()
            self._point0 = point2
        elif self._not_hemisphere and self._point0 and self._small_radius ^ v:
            t = self._intersect(point1, self._point0, True)
            if not (c & self._c0) and t:
                self._clean = 0
                if self._small_radius:
                    self._stream.line_start()
                    self._stream.point(t[0][0], t[0][1])
                    self._stream.point(t[1][0], t[1][1])
                    self._stream.line_end()
                else:
                    self._stream.point(t[1][0], t[1][1])
                    self._stream.line_end()
                    self._stream.line_start()
                    argpass(self._stream.point)(t[0][0], t[0][1], 3)
        if v and (not self._point0 or not point_equal(self._point0, point1)):
            self._stream.point(point1[0], point1[1])
        self._point0 = point1
        self._v0 = v
        self._c0 = c

    def line_end(self):
        if self._v0:
            self._stream.line_end()
        self._point0 = None

    def clean(self) -> int:
        return self._clean | ((self._v00 and self._v0) << 1)

    def _intersect(self, a: Point2D, b: Point2D, two: bool = False) -> Point2D:
        pa = cartesian(a)
        pb = cartesian(b)

        n1 = [1, 0, 0]
        n2 = cartesian_cross(pa, pb)
        n2n2 = cartesian_dot(n2, n2)
        n1n2 = n2[0]
        determinant = n2n2 - n1n2 * n1n2

        if not determinant:
            return not two and a

        c1 = self._cr * n2n2 / determinant
        c2 = -self._cr * n1n2 / determinant
        n1xn2 = cartesian_cross(n1, n2)
        A = cartesian_scale(n1, c1)
        B = cartesian_scale(n2, c2)
        cartesian_add_in_place(A, B)

        u = n1xn2
        w = cartesian_dot(A, u)
        uu = cartesian_dot(u, u)
        t2 = w * w - uu * (cartesian_dot(A, A) - 1)

        if t2 < 0:
            return

        t = sqrt(t2)
        q = cartesian_scale(u, (-w - t) / uu)
        cartesian_add_in_place(q, A)
        q = spherical(q)

        if not two:
            return q

        lambda0 = a[0]
        lambda1 = b[0]
        phi0 = a[1]
        phi1 = b[1]

        if lambda1 < lambda0:
            z = lambda0
            lambda0 = lambda1
            lambda1 = z

        delta = lambda1 - lambda0
        polar = abs(delta - pi) < EPSILON
        meridian = polar or delta < EPSILON

        if not polar and phi1 < phi0:
            z = phi0
            phi0 = phi1
            phi1 = z

        condition = (delta > pi) ^ (lambda0 <= q[0] and q[0] <= lambda1)
        if meridian:
            if polar:
                cst = phi0 if EPSILON else phi1
                condition = (phi0 + phi1 > 0) ^ (q[1] < (abs(q[0] - lambda0) < cst))
            else:
                condition = phi0 <= q[1] and q[1] <= phi1

        if condition:
            q1 = cartesian_scale(u, (-w + t) / uu)
            cartesian_add_in_place(q1, A)
            return [q, spherical(q1)]

    def _code(self, lambda_: float, phi: float) -> int:
        r = self._radius if self._small_radius else pi - self._radius
        code = 0
        if lambda_ < -r:
            code |= 1
        elif lambda_ > r:
            code |= 2
        if phi < -r:
            code |= 4
        elif phi > r:
            code |= 8
        return code


class ClipCircleWrapper:
    def __init__(self, angle: float):
        self._angle = angle
        self._cr = cos(angle)
        self._delta = radians(2)
        self._small_radius = self._cr > 0
        self._not_hemisphere = abs(self._cr) > EPSILON
        self._start = (
            [0, -self._angle] if self._small_radius else [-pi, self._angle - pi]
        )

    def clip_line(self, stream: Stream) -> ClipCircle:
        return ClipCircle(
            self._angle,
            self._cr,
            self._delta,
            self._small_radius,
            self._not_hemisphere,
            stream,
        )

    def interpolate(
        self,
        vfrom: float | None,
        vto: float | None,
        direction: float,
        stream: LineStream,
    ):
        circle_stream(stream, self._angle, self._delta, direction, vfrom, vto)

    def visible(self, lambda_: float, phi: float) -> bool:
        return cos(lambda_) * cos(phi) > self._cr

    @property
    def start(self) -> Point2D:
        return self._start


def geo_clip_circle(angle: float) -> Callable[[PolygonStream], Clip]:
    """
    Generates a clipping function which transforms a stream such that
    geometries are bounded by a small circle of radius angle around the
    projection's center. Typically used for pre-clipping.


    Parameters
    ----------
    angle : float
        Radius Angle value around the projection's center

    Returns
    -------
    Callable[[PolygonStream], Clip]
        Clipping function
    """
    clip_circle = ClipCircleWrapper(angle)
    return clip(
        clip_circle.visible,
        clip_circle.clip_line,
        clip_circle.interpolate,
        clip_circle.start,
    )



from ...types import T
from ..series import Series
from .ascending import order_ascending


def order_descending(series: list[Series[T]]) -> list[int]:
    """
    Returns a series order such that the largest series (according to the sum
    of values) is at the bottom.

    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        List of ordered indices
    """
    return order_ascending(series)[::-1]



from math import pi, sin

from ...color import rgb


def interpolate_sinebow(t):
    t = (0.5 - t) * pi

    x = sin(t)
    r = 255 * x * x

    x = sin(t + pi / 3)
    g = 255 * x * x

    x = sin(t + pi * 2 / 3)
    b = 255 * x * x
    return str(rgb(int(r), int(g), int(b)))



from math import sqrt


def symbol_cross(context, size):
    r = sqrt(size / 5) / 2
    context.move_to(-3 * r, -r)
    context.line_to(-r, -r)
    context.line_to(-r, -3 * r)
    context.line_to(r, -3 * r)
    context.line_to(r, -r)
    context.line_to(3 * r, -r)
    context.line_to(3 * r, r)
    context.line_to(r, r)
    context.line_to(r, 3 * r)
    context.line_to(-r, 3 * r)
    context.line_to(-r, r)
    context.line_to(-3 * r, r)
    context.close_path()



from math import sqrt

SQRT_3 = sqrt(3)


def symbol_triangle(context, size):
    y = -sqrt(size / (SQRT_3 * 3))
    context.move_to(0, y * 2)
    context.line_to(-SQRT_3 * y, -y)
    context.line_to(SQRT_3 * y, -y)
    context.close_path()



from __future__ import annotations

from collections.abc import Callable

from ..selection import Selection
from ..types import Accessor, Number, T
from .curves import Curve, curve_radial, curve_radial_linear
from .line import Line


class LineRadial(Line[T]):
    def __init__(self):
        Line.__init__(self)
        super().set_curve(curve_radial_linear)

    def set_curve(self, curve: Callable[[Selection], Curve]) -> LineRadial:
        """
        Sets curve.

        Parameters
        ----------
        curve : Callable[[Selection], Curve]
            Curve factory function

        Returns
        -------
        LineRadial
            Itself
        """
        return super().set_curve(curve_radial(curve))

    def angle(self, angle: Accessor[T, float] | Number) -> LineRadial:
        """
        Sets angle accessor function

        Parameters
        ----------
        angle : Accessor[T, float] | Number
            angle accessor function

        Returns
        -------
        LineRadial
            Itself
        """
        return super().x(angle)

    def radius(self, radius: Accessor[T, float] | Number) -> LineRadial:
        """
        Sets radius accessor function

        Parameters
        ----------
        radius : Accessor[T, float] | Number
            radius accessor function

        Returns
        -------
        LineRadial
            Itself
        """
        return super().y(radius)

    def get_angle(self) -> Accessor[T, float]:
        return super().get_x()

    def get_radius(self) -> Accessor[T, float]:
        return super().get_y()


def line_radial() -> LineRadial:
    """
    A radial line generator is like the Cartesian line generator except the x
    and y accessors are replaced with angle and radius accessors. Radial lines
    are positioned relative to the origin; use a transform to change the
    origin.

    Returns
    -------
    LineRadial
        Radial line generator
    """
    return LineRadial()



from __future__ import annotations

from collections.abc import Callable
from typing import Any

from ..array import argpass
from .accessors import optional
from .hierarchy import Node, compute_height


class SparseNode:
    def __init__(self):
        self.depth = -1
        self.height = 0
        self.parent = None


preroot = SparseNode()
ambiguous = {}
imputed = {}


@argpass
def default_id(d: dict[str, Any]) -> Any | None:
    return d.get("id")


@argpass
def default_parent_id(d: dict[str, Any]) -> Any | None:
    return d.get("parent_id")


class Stratify:
    """
    Stratify operator
    """

    def __init__(self):
        self._id = default_id
        self._parent_id = default_parent_id
        self._path = None

    def __call__(self, data: list[dict[str, Any]]) -> Node:
        """
        Generates a new hierarchy from the specified tabular data.

        Parameters
        ----------
        data : list[dict[str, Any]]
            Data

        Returns
        -------
        Node
            Node object
        """
        nodes = list(data)
        current_id = self._id
        current_parent_id = self._parent_id
        node_by_key = {}
        root = None

        if self._path is not None:
            norms = [normalize(self._path(d, i, data)) for i, d in enumerate(nodes)]
            parents = [parent_of(x) for x in norms]
            set_norms = set(norms)
            set_norms.add("")
            for i in parents:
                if i not in set_norms:
                    set_norms.add(i)
                    norms.append(i)
                    parents.append(parent_of(i))
                    nodes.append(imputed)

            @argpass
            def current_id(_, i):
                return norms[i]

            @argpass
            def current_parent_id(_, i):
                return parents[i]

        n = len(nodes)
        for i, d in enumerate(nodes):
            node = nodes[i] = Node(d)

            node_id = current_id(d, i, data)
            if node_id is not None:
                node_id = str(node_id)
                if node_id != "":
                    node_key = node.id = node_id
                    node_by_key[node_key] = (
                        ambiguous if node_key in node_by_key else node
                    )

            node_id = current_parent_id(d, i, data)
            if node_id is not None:
                node_id = str(node_id)
                if node_id != "":
                    node.parent = node_id

        for node in nodes:
            node_id = node.parent
            if node_id:
                parent = node_by_key.get(node_id)
                if parent == ambiguous:
                    raise ValueError("Ambiguous")
                if parent is None:
                    raise ValueError(f"Missing: {node_id!r}")
                if parent.children:
                    parent.children.append(node)
                else:
                    parent.children = [node]
                node.parent = parent
            else:
                if root is not None:
                    raise RuntimeError("Multiple roots")
                root = node

        if root is None:
            raise RuntimeError("No root")

        if self._path is not None:
            while root.data == imputed and len(root.children) == 1:
                root = root.children[0]
                n -= 1
            for node in nodes:
                if node.data != imputed:
                    break
                node.data = None

        root.parent = preroot

        def update(node):
            nonlocal n
            node.depth = node.parent.depth + 1
            n -= 1

        root.each_before(update).each_before(compute_height)
        root.parent = None
        if n > 0:
            raise RuntimeError("Cycle")
        return root

    def set_id(
        self, id_func: Callable[[Node, int, list[dict[str, Any]]], str]
    ) -> Stratify:
        """
        Sets the id accessor to the given function and returns this stratify
        operator.

        Parameters
        ----------
        id_func : Callable[[Node, int, list[dict[str, Any]]], str]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **data** (:code:`list[dict[str, Any]]`) - Data

            It should return an ID for the node's relationships in the
            conjuction with the parent ID.

        Returns
        -------
        Stratify
            Itself
        """
        self._id = argpass(optional(id_func))
        return self

    def set_parent_id(
        self, parent_id: Callable[[Node, int, list[dict[str, Any]]], str]
    ) -> Stratify:
        """
        Sets the parent id accessor to the given function and returns this
        stratify operator.

        Parameters
        ----------
        parent_id : Callable[[Node, int, list[dict[str, Any]]], str]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **data** (:code:`list[dict[str, Any]]`) - Data

            It should return an parent ID for the node's relationships in the
            conjuction with the ID.

        Returns
        -------
        Stratify
            Itself
        """
        self._parent_id = argpass(optional(parent_id))
        return self

    def set_path(
        self, path: Callable[[Node, int, list[dict[str, Any]]], str]
    ) -> Stratify:
        """
        Sets the :code:`path` accessor to the given function and returns this
        stratify operator.

        If a :code:`path` accessor is set, the :code:`id` and :code:`parent_id`
        accessors are ignored, and a unix-like hierarchy is computed on the
        slash-delimited strings returned by the path accessor, imputing parent
        nodes and ids as necessary.

        Parameters
        ----------
        path : Callable[[Node, int, list[dict[str, Any]]], str]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **data** (:code:`list[dict[str, Any]]`) - Data

            It should return the path.

        Returns
        -------
        Stratify
            Itself
        """
        self._path = argpass(optional(path))
        return self

    def get_id(self) -> Callable[[Node, int, list[dict[str, Any]]], str]:
        return self._id

    def get_parent_id(self) -> Callable[[Node, int, list[dict[str, Any]]], str]:
        return self._parent_id

    def get_path(self) -> Callable[[Node, int, list[dict[str, Any]]], str]:
        return self._path


def normalize(path: str) -> str:
    path = str(path)
    i = len(path)
    if path == "":
        return "/"
    if slash(path, i - 1) and not slash(path, i - 2):
        path = path[:-1]
    return path if path[0] == "/" else f"/{path}"


def slash(path: str, i: int) -> str:
    if path[i] == "/":
        k = 0
        i -= 1
        while i > 0 and path[i] == "\\":
            k += 1
            i -= 1
        if (k & 1) == 0:
            return True
    return False


def parent_of(path: str) -> str:
    i = len(path)
    if i < 2:
        return ""
    i -= 1
    while i > 1:
        if slash(path, i):
            break
        i -= 1
    return path[:i]


def stratify() -> Stratify:
    """
    Builds a new stratify operator with the default settings.

    Returns
    -------
    Stratify
        Stratify object
    """
    return Stratify()



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fff7bcfec44fd95f0e",
    "ffffd4fed98efe9929cc4c02",
    "ffffd4fed98efe9929d95f0e993404",
    "ffffd4fee391fec44ffe9929d95f0e993404",
    "ffffd4fee391fec44ffe9929ec7014cc4c028c2d04",
    "ffffe5fff7bcfee391fec44ffe9929ec7014cc4c028c2d04",
    "ffffe5fff7bcfee391fec44ffe9929ec7014cc4c02993404662506",
]
SCHEME_YLORBR = list(map(colors, scheme))
interpolate_ylorbr = ramp(SCHEME_YLORBR)



from ..common import Projection
from .conic_equal_area import geo_conic_equal_area


def geo_albers() -> Projection:
    """
    The Albers' equal area-conic projection. This is a U.S.-centric
    configuration of :func:`d3.geo_conic_equal_area <geo_conic_equal_area>`.

    Returns
    -------
    Projection
        Projection object
    """
    return (
        geo_conic_equal_area()
        .parallels([29.5, 45.5])
        .scale(1070)
        .translate([480, 250])
        .rotate([96, 0])
        .set_center([-0.6, 38.7])
    )



from io import StringIO

EPSILON = 1e-6


class Path:
    def __init__(self):
        self._x0 = None
        self._y0 = None
        self._x1 = None
        self._y1 = None
        self._string = StringIO()

    def move_to(self, x: float, y: float):
        self._x0 = x
        self._x1 = x
        self._y0 = y
        self._y1 = y
        self._string.write(f"M{x},{y}")

    def close_path(self):
        if self._x1 is not None:
            self._x1 = self._x0
            self._y1 = self._y0
            self._string.write("Z")

    def line_to(self, x: float, y: float):
        self._x1 = x
        self._y1 = y
        self._string.write(f"L{x},{y}")

    def arc(self, x: float, y: float, r: float, *args):
        x0 = x + r
        y0 = y

        if r < 0:
            raise ValueError("Negative radius")

        if self._x1 is None:
            self._string.write(f"M{x0},{y0}")
        elif abs(self._x1 - x0) > EPSILON or abs(self._y1 - y0) > EPSILON:
            self._string.write(f"L{x0},{y0}")
        if not r:
            return
        self._x1 = x0
        self._y1 = y0
        self._string.write(f"A{r},{r},0,1,1,{x - r},{y}A{r},{r},0,1,1,{x0},{y0}")

    def rect(self, x: float, y: float, w: float, h: float):
        self._x0 = x
        self._x1 = x
        self._y0 = y
        self._y1 = y
        self._string.write(f"M{x},{y}h{w}v{h}h{-w}Z")

    def __str__(self):
        return self._string.getvalue() or None



from .day import time_day, time_days
from .hour import time_hour, time_hours
from .millisecond import time_millisecond, time_milliseconds
from .minute import time_minute, time_minutes
from .month import time_month, time_months
from .second import time_second, time_seconds
from .ticks import time_tick_interval, time_ticks
from .week import (
    time_friday,
    time_fridays,
    time_monday,
    time_mondays,
    time_saturday,
    time_saturdays,
    time_sunday,
    time_sundays,
    time_thursday,
    time_thursdays,
    time_tuesday,
    time_tuesdays,
    time_wednesday,
    time_wednesdays,
    time_week,
    time_weeks,
)
from .year import time_year, time_years

__all__ = [
    "time_day",
    "time_days",
    "time_friday",
    "time_fridays",
    "time_hour",
    "time_hours",
    "time_millisecond",
    "time_milliseconds",
    "time_minute",
    "time_minutes",
    "time_monday",
    "time_mondays",
    "time_month",
    "time_months",
    "time_saturday",
    "time_saturdays",
    "time_second",
    "time_seconds",
    "time_sunday",
    "time_sundays",
    "time_thursday",
    "time_thursdays",
    "time_tick_interval",
    "time_ticks",
    "time_tuesday",
    "time_tuesdays",
    "time_wednesday",
    "time_wednesdays",
    "time_week",
    "time_weeks",
    "time_year",
    "time_years",
]



from .tqmt import tqmt


def ease_exp_in(t: float) -> float:
    """
    Exponential easing; raises 2 to the exponent :math:`10 \\times (t - 1)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_exp_in(0.2)
    0.002932551319648094
    """
    return tqmt(1 - t)


def ease_exp_out(t: float) -> float:
    """
    Reverse exponential easing; equivalent to :code:`1 - d3.ease_exp_in(1 -
    t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_exp_out(0.2)
    0.750733137829912
    """
    return 1 - tqmt(t)


def ease_exp_in_out(t: float) -> float:
    """
    Symmetric exponential easing; scales :func:`d3.ease_exp_in
    <detroit.ease_exp_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_exp_out <detroit.ease_exp_out>` :math:`\\forall t \\in [0.5,
    1]`. Also equivalent to :code:`d3.ease_poly.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_exp_in_out(0.2)
    0.007331378299120235
    """
    t *= 2
    if t <= 1:
        return tqmt(1 - t) * 0.5
    else:
        return (2 - tqmt(t - 1)) * 0.5



from math import sqrt


def symbol_square2(context, size):
    r = sqrt(size) * 0.4431
    context.move_to(r, r)
    context.line_to(r, -r)
    context.line_to(-r, -r)
    context.line_to(-r, r)
    context.close_path()



from math import isnan, nan, sqrt

from ...types import Point2D
from ..common import PolygonStream


class CentroidStream(PolygonStream):
    def __init__(self):
        self._point = self._centroid_point
        self._line_start = self._centroid_line_start
        self._line_end = self._centroid_line_end
        self._X0 = 0
        self._Y0 = 0
        self._Z0 = 0
        self._X1 = 0
        self._Y1 = 0
        self._Z1 = 0
        self._X2 = 0
        self._Y2 = 0
        self._Z2 = 0
        self._x00 = nan
        self._y00 = nan
        self._x0 = nan
        self._y0 = nan

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._line_start = self._centroid_ring_start
        self._line_end = self._centroid_ring_end

    def polygon_end(self):
        self._point = self._centroid_point
        self._line_start = self._centroid_line_start
        self._line_end = self._centroid_line_end

    def point(self, x: float, y: float):
        return self._point(x, y)

    def result(self) -> Point2D:
        centroid = [nan, nan]
        if self._Z2 and not isnan(self._Z2):
            centroid = [self._X2 / self._Z2, self._Y2 / self._Z2]
        elif self._Z1 and not isnan(self._Z1):
            centroid = [self._X1 / self._Z1, self._Y1 / self._Z1]
        elif self._Z0 and not isnan(self._Z0):
            centroid = [self._X0 / self._Z0, self._Y0 / self._Z0]
        self._X0 = 0
        self._Y0 = 0
        self._Z0 = 0
        self._X1 = 0
        self._Y1 = 0
        self._Z1 = 0
        self._X2 = 0
        self._Y2 = 0
        self._Z2 = 0
        return centroid

    def _centroid_point(self, x: float, y: float):
        self._X0 += x
        self._Y0 += y
        self._Z0 += 1

    def _centroid_line_start(self):
        self._point = self._centroid_point_first_line

    def _centroid_point_first_line(self, x: float, y: float):
        self._point = self._centroid_point_line
        self._x0 = x
        self._y0 = y
        self._centroid_point(x, y)

    def _centroid_point_line(self, x: float, y: float):
        dx = x - self._x0
        dy = y - self._y0
        z = sqrt(dx * dx + dy * dy)

        self._X1 += z * (self._x0 + x) * 0.5
        self._Y1 += z * (self._y0 + y) * 0.5
        self._Z1 += z

        self._x0 = x
        self._y0 = y
        self._centroid_point(x, y)

    def _centroid_line_end(self):
        self._point = self._centroid_point

    def _centroid_ring_start(self):
        self._point = self._centroid_point_first_ring

    def _centroid_ring_end(self):
        self._centroid_point_ring(self._x00, self._y00)

    def _centroid_point_first_ring(self, x: float, y: float):
        self._point = self._centroid_point_ring
        self._x00 = x
        self._x0 = x
        self._y00 = y
        self._y0 = y
        self._centroid_point(x, y)

    def _centroid_point_ring(self, x: float, y: float):
        dx = x - self._x0
        dy = y - self._y0
        z = sqrt(dx * dx + dy * dy)

        self._X1 += z * (self._x0 + x) * 0.5
        self._Y1 += z * (self._y0 + y) * 0.5
        self._Z1 += z

        z = self._y0 * x - self._x0 * y
        self._X2 += z * (self._x0 + x)
        self._Y2 += z * (self._y0 + y)
        self._Z2 += z * 3

        self._x0 = x
        self._y0 = y
        self._centroid_point(x, y)



from ...types import Point2D
from ..common import Projection
from .projection import geo_projection


class EquirectangularRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        return [lambda_, phi]

    def invert(self, lambda_: float, phi: float) -> Point2D:
        return [lambda_, phi]


def geo_equirectangular() -> Projection:
    """
    The equirectangular (plate carrée) projection

    Returns
    -------
    Projection
        Projection object
    """
    return geo_projection(EquirectangularRaw()).scale(152.63)



from __future__ import annotations

import math
from typing import overload

from ..types import Number, T
from .continuous import Transformer, copy
from .init import init_range
from .linear import LinearBase


def sign(x):
    return -1 if x < 0 else 1


def transform_symlog(c):
    return lambda x: sign(x) * math.log1p(abs(x / c))


def transform_symexp(c):
    return lambda x: sign(x) * math.expm1(abs(x)) * c


class ScaleSymlog(Transformer[float], LinearBase):
    """
    A bi-symmetric log transformation for wide-range data by Webber for
    details. Unlike a log scale, a symlog scale domain can include zero.

    Parameters
    ----------
    c : Number
        Symlog constant value
    """

    def __init__(self, c: Number = 1):
        self._c = c
        super().__init__(transform_symlog(self._c), transform_symexp(self._c))

    def set_constant(self, c: Number) -> ScaleSymlog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : Number
            Constant value

        Returns
        -------
        ScaleSymlog
            Itself
        """
        self._c = float(c)
        self._transform = transform_symlog(self._c)
        self._untransform = transform_symexp(self._c)
        self._rescale()
        return self

    def get_constant(self):
        return self._c

    def copy(self):
        return copy(self, ScaleSymlog()).set_constant(self.get_constant())


@overload
def scale_symlog() -> ScaleSymlog: ...


@overload
def scale_symlog(range_vals: list[T]) -> ScaleSymlog: ...


@overload
def scale_symlog(domain: list[Number], range_vals: list[T]) -> ScaleSymlog: ...


def scale_symlog(*args):
    """
    Builds a new continuous scale with the specified domain and
    range, the constant 1, the default interpolator and clamping
    disabled.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[Number]
        Array of values

    Returns
    -------
    ScaleSymlog
        Scale object

    Examples
    --------

    >>> scale = d3.scale_symlog([0, 100], [0, 960])
    >>> scale = scale.set_constant(2)
    >>> steps = 10
    >>> scale(0)
    0.0
    >>> scale(0.5)
    54.48303899306232
    >>> for x in range(steps + 1):
    ...     x = 2 * x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 98.9989231832065
    1.5848931924611136 142.48806851481947
    2.51188643150958 198.64193111140278
    3.9810717055349722 267.46722288814453
    6.309573444801933 347.7495648320975
    10.0 437.47847720986016
    15.848931924611133 534.4195868305859
    25.118864315095795 636.548856160817
    39.810717055349734 742.2519630838442
    63.09573444801933 850.343754278314
    """
    scale = ScaleSymlog()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e7e1efc994c7dd1c77",
    "f1eef6d7b5d8df65b0ce1256",
    "f1eef6d7b5d8df65b0dd1c77980043",
    "f1eef6d4b9dac994c7df65b0dd1c77980043",
    "f1eef6d4b9dac994c7df65b0e7298ace125691003f",
    "f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125691003f",
    "f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125698004367001f",
]
SCHEME_PURD = list(map(colors, scheme))
interpolate_purd = ramp(SCHEME_PURD)



from math import sqrt

from .enclose import Circle, pack_enclose_random


def place(b: Circle, a: Circle, c: Circle):
    dx = b.x - a.x
    dy = b.y - a.y
    d2 = dx * dx + dy * dy
    if d2:
        a2 = a.r + c.r
        a2 *= a2
        b2 = b.r + c.r
        b2 *= b2
        if a2 > b2:
            x = (d2 + b2 - a2) / (2 * d2)
            y = sqrt(max(0, b2 / d2 - x * x))
            c.x = b.x - x * dx - y * dy
            c.y = b.y - x * dy + y * dx
        else:
            x = (d2 + a2 - b2) / (2 * d2)
            y = sqrt(max(0, a2 / d2 - x * x))
            c.x = a.x + x * dx - y * dy
            c.y = a.y + x * dy + y * dx
    else:
        c.x = a.x + c.r
        c.y = a.y


def intersects(a: Circle, b: Circle) -> bool:
    dr = a.r + b.r - 1e-6
    dx = b.x - a.x
    dy = b.y - a.y
    return dr > 0 and dr * dr > dx * dx + dy * dy


class Node:
    def __init__(self, circle: Circle):
        self.circle = circle
        self.next = None
        self.previous = None


def score(node: Node) -> float:
    a = node.circle
    b = node.next.circle
    ab = a.r + b.r
    dx = (a.x * b.r + b.x * a.r) / ab
    dy = (a.y * b.r + b.y * a.r) / ab
    return dx * dx + dy * dy


def pack_siblings_random(circles: list[Circle]) -> float:
    circles = list(circles)
    n = len(circles)
    if n == 0:
        return 0

    a = circles[0]
    a.x = 0
    a.y = 0
    if n <= 1:
        return a.r

    b = circles[1]
    a.x = -b.r
    b.x = a.r
    b.y = 0

    if n <= 2:
        return a.r + b.r

    c = circles[2]
    place(b, a, c)
    a = Node(a)
    b = Node(b)
    c = Node(c)
    a.next = c.previous = b
    b.next = a.previous = c
    c.next = b.previous = a

    i = 3
    while i < n:
        continue_main = False
        c = circles[i]
        place(a.circle, b.circle, c)
        c = Node(c)

        j = b.next
        k = a.previous
        sj = b.circle.r
        sk = a.circle.r
        while True:
            if sj <= sk:
                if intersects(j.circle, c.circle):
                    b = j
                    a.next = b
                    b.previous = a
                    continue_main = True
                    break
                sj += j.circle.r
                j = j.next
            else:
                if intersects(k.circle, c.circle):
                    a = k
                    a.next = b
                    b.previous = a
                    continue_main = True
                    break
                sk += k.circle.r
                k = k.previous
            if j == k.next:
                break

        if continue_main:
            continue

        c.previous = a
        c.next = b
        a.next = b.previous = b = c

        aa = score(a)
        c = c.next
        while c != b:
            ca = score(c)
            if ca < aa:
                a = c
                aa = ca
            c = c.next
        b = a.next
        i += 1

    a = [b.circle]
    c = b

    c = c.next
    while c != b:
        a.append(c.circle)
        c = c.next
    c = pack_enclose_random(a)

    for circle in circles:
        a = circle
        a.x -= c.x
        a.y -= c.y

    return c.r


def pack_siblings(circles: list[Circle]) -> list[Circle]:
    """
    Packs the specified array of circles, each of which must have a
    :code:`circle.r` property specifying the circle's radius. Assigns the
    following properties to each circle:

    * :code:`circle.x` - the x-coordinate of the circle's center
    * :code:`circle.y` - the y coordinate of the circle's center

    The circles are positioned according to the front-chain packing algorithm
    by `Wang et al <https://dl.acm.org/doi/10.1145/1124772.1124851>`_.

    Parameters
    ----------
    circles : list[Circle]
        List of circles

    Returns
    -------
    list[Circle]
        Packed list of circles
    """
    pack_siblings_random(circles)
    return circles



def area(ring):
    n = len(ring)
    area = ring[n - 1][1] * ring[0][0] - ring[n - 1][0] * ring[0][1]
    for i in range(1, n):
        area += ring[i - 1][1] * ring[i][0] - ring[i - 1][0] * ring[i][1]
    return area



from math import cos, pi, sin

from ..types import Point2D


def point_radial(angle: float, radius: float) -> Point2D:
    """
    Returns the point :code:`[x, y]` for the given angle in radians, with 0 at
    -y (12 o'clock) and positive angles proceeding clockwise, and the given
    radius.

    Parameters
    ----------
    angle : float
        Angle value
    radius : float
        Radius value

    Returns
    -------
    Point2D
        2D Point
    """
    angle -= pi * 0.5
    return [radius * cos(angle), radius * sin(angle)]



from __future__ import annotations

# Since Javascript accepts weird stuff, I had to make the same programming
# "errors" / "mistakes" in order to keep as simple as possible the code. These
# "mistakes" are adding attributes to array object. For instance, in
# Javascript, you can do
# ```js
# var myarray = [1, 2, 3];
# myarray.data = 42
# ```
# In Python, this would raise an error. However, in this specific algorihm,
# lists have a fixed length of 4. I have used this assumption and I dynamically
# add a fifth element (of type `dict`) in `quad` objects in order to *mimic*
# the same behavior.
from collections.abc import Callable
from math import sqrt

from ..array import argpass
from ..types import (
    SimulationLink,
    SimulationLinkFunction,
    SimulationNode,
    SimulationNodeFunction,
    T,
)
from .constant import constant
from .jiggle import jiggle


def index(d: SimulationNode) -> int:
    return d["index"]


def find(node_by_id: dict[T, SimulationNode], node_id: T) -> SimulationNode:
    node = node_by_id.get(node_id)
    if node is None:
        raise RuntimeError(f"Node not found: {node_id}")
    return node


class ForceLink:
    def __init__(self, links: list[SimulationLink]):
        self._links = links
        self._id = index
        self._strength = self._default_strength
        self._strengths = None
        self._distance = constant(30)
        self._distances = None
        self._nodes = None
        self._count = None
        self._bias = None
        self._random = None
        self._iterations = 1

    def _default_strength(
        self, link: SimulationLink, i: int, links: list[SimulationLink]
    ) -> float:
        return 1 / min(
            self._count[link["source"]["index"]],
            self._count[link["target"]["index"]],
        )

    def __call__(self, alpha: float):
        for k in range(self._iterations):
            for i, link in enumerate(self._links):
                source = link["source"]
                target = link["target"]

                x = (target["x"] + target["vx"] - source["x"] - source["vx"]) or jiggle(
                    self._random
                )
                y = (target["y"] + target["vy"] - source["y"] - source["vy"]) or jiggle(
                    self._random
                )

                length = sqrt(x * x + y * y)
                length = (
                    (length - self._distances[i]) / length * alpha * self._strengths[i]
                )

                x *= length
                y *= length

                b = self._bias[i]
                target["vx"] -= x * b
                target["vy"] -= y * b

                b = 1 - b
                source["vx"] += x * b
                source["vy"] += y * b

    def _initialize(self):
        if self._nodes is None:
            return

        node_by_id = {
            self._id(node, i, self._nodes): node for i, node in enumerate(self._nodes)
        }

        self._count = [None] * len(self._nodes)
        for i, link in enumerate(self._links):
            link["index"] = i
            if not isinstance(link["source"], dict):  # ODO: check real type
                link["source"] = find(node_by_id, link["source"])
            if not isinstance(link["target"], dict):  # ODO: check real type
                link["target"] = find(node_by_id, link["target"])
            self._count[link["source"]["index"]] = (
                self._count[link["source"]["index"]] or 0
            ) + 1
            self._count[link["target"]["index"]] = (
                self._count[link["target"]["index"]] or 0
            ) + 1

        self._bias = [None] * len(self._links)
        for i, link in enumerate(self._links):
            self._bias[i] = self._count[link["source"]["index"]] / (
                self._count[link["source"]["index"]]
                + self._count[link["target"]["index"]]
            )

        self._strengths = [None] * len(self._links)
        self._initialize_strength()
        self._distances = [None] * len(self._links)
        self._initialize_distance()

    def _initialize_strength(self):
        if self._nodes is None:
            return

        for i, link in enumerate(self._links):
            self._strengths[i] = self._strength(link, i, self._links)

    def _initialize_distance(self):
        if self._nodes is None:
            return

        for i, link in enumerate(self._links):
            self._distances[i] = self._distance(link, i, self._links)

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes
        self._random = random
        self._initialize()

    def set_links(self, links: list[SimulationLink]) -> ForceLink:
        """
        Sets the array of links associated with this force, recomputes the
        distance and strength parameters for each link, and returns this force.

        Each link is an object with the following properties:

        * **source** - the link's source node; see simulation.nodes
        * **target** - the link's target node; see simulation.nodes
        * **index** - the zero-based index into links, assigned by this method

        For convenience, a link's source and target properties may be
        initialized using numeric or string identifiers rather than object
        references.

        Parameters
        ----------
        links : list[SimulationLink]
            List of links

        Returns
        -------
        ForceLink
            Itself
        """
        self._links = links
        self._initialize()
        return self

    def set_id(self, id_func: SimulationNodeFunction[int]) -> ForceLink:
        """
        Sets the node ID accessor to the specified function and returns this
        force.

        Parameters
        ----------
        id_func : SimulationNodeFunction[int]
            ID accessor function which takes the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the ID value (:code:`int`)

        Returns
        -------
        ForceLink
            Itself
        """
        self._id = argpass(id_func)
        return self

    def set_iterations(self, iterations: int) -> ForceLink:
        """
        Sets the number of iterations per application to the specified number
        and returns this force.

        Parameters
        ----------
        iterations : int
            Number of iterations

        Returns
        -------
        ForceLink
            Itself
        """
        self._iterations = iterations
        return self

    def set_strength(
        self, strength: SimulationLinkFunction[float] | float
    ) -> ForceLink:
        """
        Sets the strength accessor to the specified number or function,
        re-evaluates the strength accessor for each link, and returns this
        force.

        Parameters
        ----------
        strength : SimulationLinkFunction[float] | float
            Strength function or constant value. If it is a function, it takes
            the following arguments:

            * **link** (:code:`SimulationLink`) - the link element
            * **i** (:code:`int`) - the index of the node
            * **links** (:code:`list[SimulationLink]`) - the list of links

            It returns the strength value (:code:`float`)

        Returns
        -------
        ForceLink
            Itself
        """
        if callable(strength):
            self._strength = argpass(strength)
        else:
            self._strength = constant(strength)
        self._initialize_strength()
        return self

    def set_distance(
        self, distance: SimulationNodeFunction[float] | float
    ) -> ForceLink:
        """
        Sets the distance accessor to the specified number or function,
        re-evaluates the distance accessor for each link, and returns this
        force.

        Parameters
        ----------
        distance : SimulationNodeFunction[float] | float
            Distance function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the distance value (:code:`float`)

        Returns
        -------
        ForceLink
            Itself
        """
        if callable(distance):
            self._distance = argpass(distance)
        else:
            self._distance = constant(distance)
        self._initialize_distance()
        return self

    def get_links(self) -> list[SimulationLink]:
        return self._links

    def get_id(self) -> SimulationNodeFunction[int]:
        return self._id

    def get_iterations(self) -> int:
        return self._iterations

    def get_strength(self) -> SimulationNodeFunction[float]:
        return self._strength

    def get_distance(self) -> SimulationNodeFunction[float]:
        return self._distance


def force_link(links: list[SimulationLink] | None = None) -> ForceLink:
    """
    The link force pushes linked nodes together or apart according to the
    desired link distance. The strength of the force is proportional to the
    difference between the linked nodes' distance and the target distance,
    similar to a spring force.

    Parameters
    ----------
    links : list[SimulationLink] | None
        List of links

    Returns
    -------
    ForceLink
        Force object
    """
    if links is None:
        links = []
    return ForceLink(links)



from __future__ import annotations

from collections.abc import Callable

from ..types import SimulationNode


class ForceCenter:
    def __init__(self, x: float, y: float):
        self._x = x
        self._y = y
        self._strength = 1
        self._nodes = None

    def __call__(self, alpha: float | None = None):
        n = len(self._nodes)
        sx = 0
        sy = 0

        for node in self._nodes:
            sx += node["x"]
            sy += node["y"]

        sx = sx / n - self._x * self._strength
        sy = sy / n - self._y * self._strength

        for node in self._nodes:
            node["x"] -= sx
            node["y"] -= sy

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes

    def x(self, x: float) -> ForceCenter:
        """
        Sets the x-coordinate of the centering position to the specified number
        and returns this force.

        Parameters
        ----------
        x : float
            X-coordinate of the centering position

        Returns
        -------
        ForceCenter
            Itself
        """
        self._x = x
        return self

    def y(self, y: float) -> ForceCenter:
        """
        Sets the y-coordinate of the centering position to the specified number
        and returns this force.

        Parameters
        ----------
        y : float
            Y-coordinate of the centering position

        Returns
        -------
        ForceCenter
            Itself
        """
        self._y = y
        return self

    def set_strength(self, strength: float) -> ForceCenter:
        """
        Sets the center force's strength. A reduced strength of e.g. 0.05
        softens the movements on interactive graphs in which new nodes enter or
        exit the graph.

        Parameters
        ----------
        strength : float
            Strength value

        Returns
        -------
        ForceCenter
            Itself
        """
        self._strength = strength
        return self

    def get_x(self):
        return self._x

    def get_y(self):
        return self._y

    def get_strength(self) -> float:
        return self._strength


def force_center(x: float | None = None, y: float | None = None) -> ForceCenter:
    """
    The center force translates nodes uniformly so that the mean position of
    all nodes (the center of mass if all nodes have equal weight) is at the
    given position :math:`(x,y)`. This force modifies the positions of nodes on
    each application; it does not modify velocities, as doing so would
    typically cause the nodes to overshoot and oscillate around the desired
    center. This force helps keep nodes in the center of the viewport, and
    unlike the position forces, it does not distort their relative positions.

    Parameters
    ----------
    x : float | None
        X-coordinate of the centering position
    y : float | None
        Y-coordinate of the centering position

    Returns
    -------
    ForceCenter
        Force object
    """
    if x is None:
        x = 0
    if y is None:
        y = 0
    return ForceCenter(x, y)



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Generic

from ..array import argpass
from ..selection.selection import Selection
from ..types import Accessor, Number, T
from .constant import constant
from .curves import Curve, curve_linear
from .line import Line
from .path import WithPath
from .point import x as point_x
from .point import y as point_y


class Area(Generic[T], WithPath):
    """
    The area generator produces an area defined by a topline and a baseline as in
    an area chart. Typically, the two lines share the same x-values (x0 = x1),
    differing only in y-value (y0 and y1); most commonly, y0 is defined as a
    constant representing zero (the y scale's output for zero). The topline
    is defined by x1 and y1 and is rendered first; the baseline is defined by
    x0 and y0 and is rendered second with the points in reverse order. With a
    curveLinear curve, this produces a clockwise polygon. See also radial areas.

    Parameters
    ----------
    x0 : Accessor[T, float] | None
        x0 accessor function for data points
    y0 : Accessor[T, float] | None
        y0 accessor function for data points
    y1 : Accessor[T, float] | None
        y1 accessor function for data points

    Returns
    -------
    Area
        New area generator
    """

    def __init__(
        self,
        x0: Accessor[T, float] | None = None,
        y0: Accessor[T, float] | None = None,
        y1: Accessor[T, float] | None = None,
    ):
        super().__init__()
        self._x1 = None
        self._defined = constant(True)
        self._context = None
        self._curve = curve_linear
        self._output = None

        if x0 is None:
            self._x0 = point_x
        elif callable(x0):
            self._x0 = x0
        else:
            self._x0 = constant(x0)

        if y0 is None:
            self._y0 = constant(0)
        elif callable(y0):
            self._y0 = y0
        else:
            self._y0 = constant(y0)

        if y1 is None:
            self._y1 = point_y
        elif callable(y1):
            self._y1 = y1
        else:
            self._y1 = constant(y1)

        self._x0 = argpass(self._x0)
        self._x1 = argpass(self._x1) if self._x1 else None
        self._y0 = argpass(self._y0)
        self._y1 = argpass(self._y1) if self._y1 else None

    def __call__(self, data: Iterable[T]) -> str | None:
        """
        Generates an area for the given array of data.

        Parameters
        ----------
        data : Iterable[T]
            Data values

        Returns
        -------
        str | None
            Generated area if the area is not associated to a context

        Examples
        --------

        >>> svg.append("path").attr("d", area(data))
        """
        data = list(data)
        n = len(data)
        if n == 0:
            return None
        defined0 = False

        x0z = [0] * n
        y0z = [0] * n

        buffer = None
        if self._context is None:
            buffer = self._path()
            self._output = self._curve(buffer)

        j = 0
        for i in range(n + 1):
            d = data[i] if i < n else None
            if not (i < n and self._defined(d, i, data) == defined0):
                defined0 = not defined0
                if defined0:
                    j = i
                    self._output.area_start()
                    self._output.line_start()
                else:
                    self._output.line_end()
                    self._output.line_start()
                    for k in range(i - 1, j - 1, -1):
                        self._output.point(x0z[k], y0z[k])
                    self._output.line_end()
                    self._output.area_end()

            if defined0:
                x0z[i] = self._x0(d, i, data)
                y0z[i] = self._y0(d, i, data)
                self._output.point(
                    self._x1(d, i, data) if self._x1 else x0z[i],
                    self._y1(d, i, data) if self._y1 else y0z[i],
                )

        if buffer:
            self._output = None
            return str(buffer) or None

    def area_line(self) -> Line:
        """
        Returns a new line generator from the definition of the area.

        Returns
        -------
        Line
            Line generator based on the area parameters
        """
        return (
            Line()
            .set_defined(self._defined)
            .set_curve(self._curve)
            .set_context(self._context)
        )

    def line_x0(self) -> Line:
        """
        Returns a new line generator that has this area generator's current
        defined accessor, curve and context. The line's x-accessor is this
        area's x0-accessor, and the line's y-accessor is this area's
        y0-accessor.

        Returns
        -------
        Line
            Line generator based on the area parameters
        """
        return self.area_line().x(self._x0).y(self._y0)

    def line_y0(self) -> Line:
        """
        Returns a new line generator that has this area generator's current
        defined accessor, curve and context. The line's x-accessor is this
        area's x0-accessor, and the line's y-accessor is this area's
        y0-accessor.

        Returns
        -------
        Line
            Line generator based on the area parameters
        """
        return self.area_line().x(self._x0).y(self._y0)

    def line_x1(self) -> Line:
        """
        Returns a new line generator that has this area generator's current
        defined accessor, curve and context. The line's x-accessor is this
        area's x1-accessor, and the line's y-accessor is this area's
        y0-accessor.

        Returns
        -------
        Line
            Line generator based on the area parameters
        """
        return self.area_line().x(self._x1).y(self._y0)

    def line_y1(self) -> Line:
        """
        Returns a new line generator that has this area generator's current
        defined accessor, curve and context. The line's x-accessor is this
        area's x0-accessor, and the line's y-accessor is this area's
        y1-accessor.

        Returns
        -------
        Line
            Line generator based on the area parameters
        """
        return self.area_line().x(self._x0).y(self._y1)

    def x(self, x: Accessor[T, float] | Number) -> Area:
        """
        Sets x0 accessor function and sets x1
        accessor function to :code:`None`

        Parameters
        ----------
        x : Accessor[T, float] | Number
            x accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(x):
            self._x0 = x
        else:
            self._x0 = constant(x)
        self._x0 = argpass(self._x0)
        self._x1 = None
        return self

    def x0(self, x0: Accessor[T, float] | Number) -> Area:
        """
        Sets x0 accessor function

        Parameters
        ----------
        x0 : Accessor[T, float] | Number
            x0 accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(x0):
            self._x0 = x0
        else:
            self._x0 = constant(x0)
        self._x0 = argpass(self._x0)
        return self

    def x1(self, x1: Accessor[T, float] | Number) -> Area:
        """
        Sets x1 accessor function

        Parameters
        ----------
        x1 : Accessor[T, float] | Number
            x1 accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(x1):
            self._x1 = x1
        else:
            self._x1 = constant(x1)
        self._x1 = argpass(self._x1)
        return self

    def y(self, y: Accessor[T, float] | Number) -> Area:
        """
        Sets y0 accessor function and sets y1
        accessor function to :code:`None`

        Parameters
        ----------
        y : Accessor[T, float] | Number
            y accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(y):
            self._y0 = y
        else:
            self._y0 = constant(y)
        self._y0 = argpass(self._y0)
        self._y1 = None
        return self

    def y0(self, y0: Accessor[T, float] | Number) -> Area:
        """
        Sets y accessor function

        Parameters
        ----------
        y0 : Accessor[T, float] | Number
            y0 accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(y0):
            self._y0 = y0
        else:
            self._y0 = constant(y0)
        self._y0 = argpass(self._y0)
        return self

    def y1(self, y1: Accessor[T, float] | Number) -> Area:
        """
        Sets y accessor function

        Parameters
        ----------
        y1 : Accessor[T, float] | Number
            y1 accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(y1):
            self._y1 = y1
        else:
            self._y1 = constant(y1)
        self._y1 = argpass(self._y1)
        return self

    def set_defined(self, defined: Accessor[T, bool] | Number) -> Area:
        """
        Sets defined accessor

        When an area is generated, the defined accessor will be invoked
        for each element in the input data array, being passed the element
        :code:`d`, the index :code:`i`, and the array :code:`data` as three
        arguments. If the given element is defined (i.e., if the defined
        accessor returns a truthy value for this element), the :code:`x0`,
        :code:`x1`, :code:`y0` and :code:`y1` accessors will subsequently
        be evaluated and the point will be added to the current area segment.
        Otherwise, the element will be skipped, the current area segment will
        be ended, and a new area segment will be generated for the next
        defined point. As a result, the generated area may have several
        discrete segments.


        Parameters
        ----------
        defined : Accessor[T, bool] | Number
            defined accessor function

        Returns
        -------
        Area
            Itself
        """
        if callable(defined):
            self._defined = defined
        else:
            self._defined = constant(bool(defined))
        return self

    def set_curve(self, curve: Callable[[Selection], Curve] | None = None) -> Area:
        """
        Sets curve factory.

        Parameters
        ----------
        curve : Callable[[Selection], Curve] | None
            Curve factory function

        Returns
        -------
        Area
            Itself
        """
        self._curve = curve
        if self._context is not None:
            self._output = self._curve(self._context)
        return self

    def set_context(self, context: Selection | None = None) -> Area:
        """
        Sets the context.

        Parameters
        ----------
        context : Selection | None
            Selection

        Returns
        -------
        Area
            Itself
        """
        if context is None:
            self._context = None
            self._output = None
        else:
            self._context = context
            self._output = self._curve(self._context)
        return self

    def get_x(self) -> Accessor[T, float]:
        return self._x0

    def get_y(self) -> Accessor[T, float]:
        return self._y0

    def get_x0(self) -> Accessor[T, float]:
        return self._x0

    def get_x1(self) -> Accessor[T, float]:
        return self._x1

    def get_y0(self) -> Accessor[T, float]:
        return self._y0

    def get_y1(self) -> Accessor[T, float]:
        return self._y1

    def get_defined(self) -> Accessor[T, float]:
        return self._defined

    def get_curve(self) -> Callable[[Selection], Curve]:
        return self._curve

    def get_context(self) -> Selection:
        return self._context



from .chord import chord, chord_directed, chord_transpose
from .ribbon import ribbon, ribbon_arrow

__all__ = [
    "chord",
    "chord_directed",
    "chord_transpose",
    "ribbon",
    "ribbon_arrow",
]



from math import asin, atan2, cos, fsum, pi, sin

from ..types import Point2D
from .cartesian import cartesian, cartesian_cross, cartesian_normalize_in_place

EPSILON = 1e-6
EPSILON2 = 1e-12
TAU = 2 * pi
half_pi = pi * 0.5
quarter_pi = pi * 0.25


def sign(x: float) -> float:
    return -1 if x < 0 else 1


def longitude(point: Point2D) -> float:
    if abs(point[0]) <= pi:
        return point[0]
    else:
        return sign(point[0]) * ((abs(point[0]) + pi) % TAU - pi)


def polygon_contains(polygon: list[list[float]], point: tuple[float, float]) -> int:
    lambda_ = longitude(point)
    phi = point[1]
    sin_phi = sin(phi)
    normal = [sin(lambda_), -cos(lambda_), 0]
    angle = 0
    winding = 0

    sum = []

    if sin_phi == 1.0:
        phi = half_pi + EPSILON
    elif sin_phi == -1.0:
        phi = -half_pi - EPSILON

    for ring in polygon:
        m = len(ring)
        if m == 0:
            continue
        point0 = ring[m - 1]
        lambda0 = longitude(point0)
        phi0 = point0[1] * 0.5 + quarter_pi
        sin_phi0 = sin(phi0)
        cos_phi0 = cos(phi0)

        for j in range(m):
            point1 = ring[j]
            lambda1 = longitude(point1)
            phi1 = point1[1] * 0.5 + quarter_pi
            sin_phi1 = sin(phi1)
            cos_phi1 = cos(phi1)
            delta = lambda1 - lambda0
            sign_ = sign(delta)
            abs_delta = sign_ * delta
            antimeridian = abs_delta > pi
            k = sin_phi0 * sin_phi1

            sum.append(
                atan2(
                    k * sign_ * sin(abs_delta), cos_phi0 * cos_phi1 + k * cos(abs_delta)
                )
            )
            angle += delta + sign_ * TAU if antimeridian else delta

            if (antimeridian ^ (lambda0 >= lambda_)) ^ (lambda1 >= lambda_):
                arc = cartesian_cross(cartesian(point0), cartesian(point1))
                cartesian_normalize_in_place(arc)
                intersection = cartesian_cross(normal, arc)
                cartesian_normalize_in_place(intersection)
                phi_arc = (-1 if antimeridian ^ (delta >= 0) else 1) * asin(
                    intersection[2]
                )
                if phi > phi_arc or phi == phi_arc and (arc[0] or arc[1]):
                    winding += 1 if antimeridian ^ (delta >= 0) else -1

            lambda0 = lambda1
            sin_phi0 = sin_phi1
            cos_phi0 = cos_phi1
            point0 = point1

    return (angle < -EPSILON or angle < EPSILON and fsum(sum) < -EPSILON2) ^ (
        winding & 1
    )



from functools import cmp_to_key
from math import degrees, fsum, inf, isinf, nan, radians

from ..types import GeoJSON, Point2D
from .area import AreaStream
from .cartesian import (
    cartesian,
    cartesian_cross,
    cartesian_normalize_in_place,
    spherical,
)
from .common import PolygonStream
from .stream import geo_stream

EPSILON = 1e-6


def angle(lambda0: float, lambda1: float) -> float:
    lambda1 -= lambda0
    return lambda1 + 360 if lambda1 < 0 else lambda1


def range_compare(a: Point2D, b: Point2D) -> float:
    return a[0] - b[0]


def range_contains(rang: Point2D, x: float) -> bool:
    if rang[0] <= rang[1]:
        return rang[0] <= x <= rang[1]
    else:
        return x < rang[0] or rang[1] < x


class BoundsStream(PolygonStream):
    def __init__(self):
        self._lambda0 = inf
        self._phi0 = inf
        self._lambda1 = -inf
        self._phi1 = -inf

        self._lambda2 = None

        self._lambda00 = None
        self._phi00 = None

        self._p0 = None
        self._delta_sum = []
        self._ranges = []
        self._range = [nan, nan]

        self._point = self._bounds_point
        self._line_start = self._bounds_line_start
        self._line_end = self._bounds_line_end

        self._area_stream = AreaStream()

    def point(self, x: float, y: float):
        return self._point(x, y)

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._point = self._bounds_ring_point
        self._line_start = self._bounds_ring_start
        self._line_end = self._bounds_ring_end
        self._delta_sum = []
        self._area_stream.polygon_start()

    def polygon_end(self):
        self._area_stream.polygon_end()
        self._point = self._bounds_point
        self._line_start = self._bounds_line_start
        self._line_end = self._bounds_line_end
        area_ring_sum = fsum(self._area_stream._area_ring_sum)
        delta_sum = fsum(self._delta_sum)
        if area_ring_sum < 0:
            self._lambda1 = 180
            self._lambda0 = -180
            self._phi1 = 90
            self._phi0 = -90
        elif delta_sum > EPSILON:
            self._phi1 = 90
        elif delta_sum < -EPSILON:
            self._phi0 = -90
        self._range[0] = self._lambda0
        self._range[1] = self._lambda1

    def sphere(self):
        self._lambda1 = 180
        self._lambda0 = -180
        self._phi1 = 90
        self._phi0 = -90

    def _bounds_point(self, lambda_: float, phi: float):
        self._lambda0 = lambda_
        self._lambda1 = lambda_
        self._range = [self._lambda0, self._lambda1]
        self._ranges.append(self._range)
        if phi < self._phi0:
            self._phi0 = phi
        if phi > self._phi1:
            self._phi1 = phi

    def _line_point(self, lambda_: float, phi: float):
        p = cartesian([radians(lambda_), radians(phi)])
        if self._p0:
            normal = cartesian_cross(self._p0, p)
            equatorial = [normal[1], -normal[0], 0]
            inflection = cartesian_cross(equatorial, normal)
            cartesian_normalize_in_place(inflection)
            inflection = spherical(inflection)
            delta = lambda_ - self._lambda2
            sign = 1 if delta > 0 else -1
            lambdai = degrees(inflection[0]) * sign
            antimeridian = abs(delta) > 180

            if antimeridian ^ (
                sign * self._lambda2 < lambdai and lambdai < sign * lambda_
            ):
                phii = degrees(inflection[1])
                if phii > self._phi1:
                    self._phi1 = phii
            else:
                lambdai = (lambdai + 360) % 360 - 180
                if antimeridian ^ (
                    sign * self._lambda2 < lambdai and lambdai < sign * lambda_
                ):
                    phii = degrees(-inflection[1])
                    if phii < self._phi0:
                        self._phi0 = phii
                else:
                    if phi < self._phi0:
                        self._phi0 = phi
                    if phi > self._phi1:
                        self._phi1 = phi

            if antimeridian:
                if lambda_ < self._lambda2:
                    if angle(self._lambda0, lambda_) > angle(
                        self._lambda0, self._lambda1
                    ):
                        self._lambda1 = lambda_
                else:
                    if angle(lambda_, self._lambda1) > angle(
                        self._lambda0, self._lambda1
                    ):
                        self._lambda0 = lambda_
            else:
                if self._lambda1 >= self._lambda0:
                    if lambda_ < self._lambda0:
                        self._lambda0 = lambda_
                    if lambda_ > self._lambda1:
                        self._lambda1 = lambda_
                else:
                    if lambda_ > self._lambda2:
                        if angle(self._lambda0, lambda_) > angle(
                            self._lambda0, self._lambda1
                        ):
                            self._lambda1 = lambda_
                    else:
                        if angle(lambda_, self._lambda1) > angle(
                            self._lambda0, self._lambda1
                        ):
                            self._lambda0 = lambda_
        else:
            self._lambda0 = lambda_
            self._lambda1 = lambda_
            self._range = [self._lambda0, self._lambda1]
            self._ranges.append(self._range)

        if phi < self._phi0:
            self._phi0 = phi
        if phi > self._phi1:
            self._phi1 = phi
        self._p0 = p
        self._lambda2 = lambda_

    def _bounds_line_start(self):
        self._point = self._line_point

    def _bounds_line_end(self):
        self._range[0] = self._lambda0
        self._range[1] = self._lambda1
        self._point = self._bounds_point
        self._p0 = None

    def _bounds_ring_point(self, lambda_: float, phi: float):
        if self._p0:
            delta = lambda_ - self._lambda2
            if abs(delta) > 180:
                delta = delta + (360 if delta > 0 else -360)
            self._delta_sum.append(delta)
        else:
            self._lambda00 = lambda_
            self._phi00 = phi
        self._area_stream.point(lambda_, phi)
        self._line_point(lambda_, phi)

    def _bounds_ring_start(self):
        self._area_stream.line_start()

    def _bounds_ring_end(self):
        self._bounds_ring_point(self._lambda00, self._phi00)
        self._area_stream.line_end()
        delta_sum = fsum(self._delta_sum)
        if abs(delta_sum) > EPSILON:
            self._lambda1 = 180
            self._lambda0 = -180
        self._range[0] = self._lambda0
        self._range[1] = self._lambda1
        self._p0 = None

    def result(self) -> tuple[Point2D, Point2D]:
        if n := len(self._ranges):
            self._ranges.sort(key=cmp_to_key(range_compare))
            a = self._ranges[0]
            merged = [a]
            for i in range(1, n):
                b = self._ranges[i]
                if range_contains(a, b[0]) or range_contains(a, b[1]):
                    if angle(a[0], b[1]) > angle(a[0], a[1]):
                        a[1] = b[1]
                    if angle(b[0], a[1]) > angle(a[0], a[1]):
                        a[0] = b[0]
                else:
                    a = b
                    merged.append(a)

            delta_max = -inf
            n = len(merged)
            a = merged[n - 1]
            for i in range(n):
                b = merged[i]
                delta = angle(a[1], b[0])
                if delta > delta_max:
                    delta_max = delta
                    self._lambda0 = b[0]
                    self._lambda1 = a[1]
                a = b

        self._ranges = []
        self._range = []
        if isinf(self._lambda0) or isinf(self._phi0):
            return [[nan, nan], [nan, nan]]
        else:
            return [[self._lambda0, self._phi0], [self._lambda1, self._phi1]]


def geo_bounds(obj: GeoJSON) -> tuple[Point2D, Point2D]:
    """
    Returns the spherical bounding box for the specified GeoJSON object. The
    bounding box is represented by a two-dimensional array: :code:`[[left,
    bottom], [right, top]]`, where :code:`left` is the minimum longitude,
    :code:`bottom` is the minimum latitude, :code:`right` is maximum longitude,
    and :code:`top` is the maximum latitude. All coordinates are given in
    degrees. (Note that in projected planar coordinates, the minimum latitude
    is typically the maximum y-value, and the maximum latitude is typically the
    minimum y-value.) This is the spherical equivalent of :func:`GeoPath.bounds
    <detroit.geo.path.path.GeoPath.bounds>`.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object

    Returns
    -------
    tuple[Point2D, Point2D]
        Spherical bounding box
    """
    stream = BoundsStream()
    geo_stream(obj, stream)
    return stream.result()



from collections.abc import Callable

from ..types import ContinuousScaler, SequentialScaler, U, V


def init_range(
    obj: ContinuousScaler,
    domain: list[U] | None = None,
    range_vals: list[V] | None = None,
):
    if domain is None and range_vals is None:
        return obj
    elif domain is None:
        return obj.set_range(range_vals)
    else:
        return obj.set_range(range_vals).set_domain(domain)


def init_interpolator(
    obj: SequentialScaler,
    domain: list[U] | None = None,
    interpolator: Callable[[V], V] | None = None,
):
    if domain is None and interpolator is None:
        return obj
    elif domain is None:
        if callable(interpolator):
            obj.set_interpolator(interpolator)
        else:
            obj.set_range(interpolator)
    else:
        obj.set_domain(domain)
        if callable(interpolator):
            obj.set_interpolator(interpolator)
        else:
            obj.set_range(domain)
    return obj



from collections.abc import Callable


def basis(t1, v0, v1, v2, v3):
    t2 = t1 * t1
    t3 = t2 * t1
    return (
        (1 - 3 * t1 + 3 * t2 - t3) * v0
        + (4 - 6 * t2 + 3 * t3) * v1
        + (1 + 3 * t1 + 3 * t2 - 3 * t3) * v2
        + t3 * v3
    ) / 6


def interpolate_basis(values: list[float]) -> Callable[[float], float]:
    """
    Returns a uniform nonrational B-spline interpolator
    through the specified array of values, which must be numbers.

    Implicit control points are generated such that the interpolator
    returns :code:`values[0]` at :math:`t = 0`
    and :code:`values[-1]` at :math:`t = 1`.

    Parameters
    ----------
    values : list[float]
        List of inputs

    Returns
    -------
    Callable[[float], float]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_basis([0, 10, 20])
    >>> interpolator(0)
    0.0
    >>> interpolator(1)
    20.0
    >>> interpolator(0.5)
    10.0
    """
    n = len(values) - 1

    def interpolate(t):
        i = 0 if t <= 0 else (n - 1 if t >= 1 else int(t * n))
        v1 = values[i]
        v2 = values[i + 1]
        v0 = values[i - 1] if i > 0 else 2 * v1 - v2
        v3 = values[i + 2] if i < n - 1 else 2 * v2 - v1
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate



def ease_quad_in(t: float) -> float:
    """
    Quadratic easing; equivalent to :code:`d3.ease_poly_in.exponent(2)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_quad_in(0.2)
    0.04000000000000001
    """
    return t * t


def ease_quad_out(t: float) -> float:
    """
    Reverse quadratic easing; equivalent to :code:`1 - d3.ease_quad_in(1 - t)`.
    Also equivalent to :code:`d3.ease_poly_out.exponent(2)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_quad_out(0.2)
    0.36000000000000004
    """
    return t * (2 - t)


def ease_quad_in_out(t: float) -> float:
    """
    Symmetric quadratic easing; scales :func:`d3.ease_quad_in
    <detroit.ease_quad_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_quad_out <detroit.ease_quad_out>` :math:`\\forall t \\in
    [0.5, 1]`. Also equivalent to :code:`d3.ease_poly.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_quad_in_out(0.2)
    0.08000000000000002
    """
    t *= 2
    if t <= 1:
        return t * t * 0.5
    else:
        t -= 1
        return (t * (2 - t) + 1) * 0.5



from __future__ import annotations

from collections.abc import Callable
from math import acos, asin, atan2, cos, nan, pi, sin, sqrt

from ..selection.selection import Selection
from .constant import constant
from .path import WithPath

EPSILON = 1e-12


def arc_inner_radius(d, *args):
    return d.get("inner_radius", 0)


def arc_outer_radius(d, *args):
    return d.get("outer_radius", 0)


def arc_start_angle(d, *args):
    return d.get("start_angle", 0)


def arc_end_angle(d, *args):
    return d.get("end_angle", 0)


def arc_pad_angle(d={}, *args):
    return d.get("pad_angle", 0)


def intersect(x0, y0, x1, y1, x2, y2, x3, y3):
    x10 = x1 - x0
    y10 = y1 - y0
    x32 = x3 - x2
    y32 = y3 - y2
    t = y32 * x10 - x32 * y10
    if t * t < EPSILON:
        return
    t = (x32 * (y0 - y2) - y32 * (x0 - x2)) / t
    return x0 + t * x10, y0 + t * y10


class CornerTangents:
    def __init__(self, x0, y0, x1, y1, r1, rc, cw):
        x01 = x0 - x1
        y01 = y0 - y1
        lo = (rc if cw else -rc) / sqrt(x01 * x01 + y01 * y01)
        ox = lo * y01
        oy = -lo * x01
        x11 = x0 + ox
        y11 = y0 + oy
        x10 = x1 + ox
        y10 = y1 + oy
        x00 = (x11 + x10) / 2
        y00 = (y11 + y10) / 2
        dx = x10 - x11
        dy = y10 - y11
        d2 = dx * dx + dy * dy
        r = r1 - rc
        D = x11 * y10 - x10 * y11
        d = (-1 if dy < 0 else 1) * sqrt(max(0, r * r * d2 - D * D))
        cx0 = (D * dy - dx * d) / d2
        cy0 = (-D * dx - dy * d) / d2
        cx1 = (D * dy + dx * d) / d2
        cy1 = (-D * dx + dy * d) / d2
        dx0 = cx0 - x00
        dy0 = cy0 - y00
        dx1 = cx1 - x00
        dy1 = cy1 - y00

        if dx0 * dx0 + dy0 * dy0 > dx1 * dx1 + dy1 * dy1:
            cx0 = cx1
            cy0 = cy1

        self.cx = cx0
        self.cy = cy0
        self.x01 = -ox
        self.y01 = -oy
        self.x11 = cx0 * (r1 / r - 1)
        self.y11 = cy0 * (r1 / r - 1)


class Arc(WithPath):
    """
    The arc generator produces a circular or annular sector,
    as in a pie or donut chart. Arcs are centered at the origin;
    use a transform to move the arc to a different position

    Examples
    --------

    >>> from math import pi
    >>> arc = d3.arc()
    ...     .set_inner_radius(0)
    ...     .set_outer_radius(100)
    ...     .set_start_angle(0)
    ...     .set_end_angle(pi / 2)
    """

    def __init__(self):
        super().__init__()
        self._inner_radius = arc_inner_radius
        self._outer_radius = arc_outer_radius
        self._corner_radius = constant(0)
        self._pad_radius = None
        self._start_angle = arc_start_angle
        self._end_angle = arc_end_angle
        self._pad_angle = arc_pad_angle
        self._context = None

    def __call__(self, *args) -> str | None:
        """
        Generates an arc for the given arguments. The arguments
        are arbitrary; they are propagated to the arc generator's
        accessor functions along with the this object.

        Parameters
        ----------
        args : Any
            Additional arguments

        Returns
        -------
        str | None
            Generated arc if the arc is not associated to a context

        Examples
        --------

        >>> from math import pi
        >>> d3.arc()
        ...     .set_inner_radius(0)
        ...     .set_outer_radius(100)
        ...     .set_start_angle(0)
        ...     .set_end_angle(pi / 2)()
        """
        r0 = self._inner_radius(*args)
        r1 = self._outer_radius(*args)
        a0 = self._start_angle(*args) - pi * 0.5
        a1 = self._end_angle(*args) - pi * 0.5
        da = abs(a1 - a0)
        cw = a1 > a0

        buffer = None
        if self._context is None:
            self._context = buffer = self._path()

        if r1 < r0:
            r = r1
            r1 = r0
            r0 = r

        if r1 <= EPSILON:
            self._context.move_to(0, 0)
        elif da > 2 * pi - EPSILON:
            self._context.move_to(r1 * cos(a0), r1 * sin(a0))
            self._context.arc(0, 0, r1, a0, a1, not cw)
            if r0 > EPSILON:
                self._context.move_to(r0 * cos(a1), r0 * sin(a1))
                self._context.arc(0, 0, r0, a1, a0, cw)
        else:
            a01 = a0
            a11 = a1
            a00 = a0
            a10 = a1
            da0 = da
            da1 = da
            ap = self._pad_angle(*args) / 2
            rp = (
                self._pad_radius(*args)
                if self._pad_radius
                else sqrt(r0 * r0 + r1 * r1)
                if ap > EPSILON
                else 0
            )
            rc = min(abs(r1 - r0) / 2, self._corner_radius(*args))
            rc0 = rc
            rc1 = rc

            if rp > EPSILON:
                tmp = rp / r0 * sin(ap) if r0 != 0 else nan
                p0 = asin(tmp) if 0 <= tmp <= 1 else nan
                p1 = asin(rp / r1 * sin(ap))

                da0 -= p0 * 2
                if da0 > EPSILON:
                    p0 *= 1 if cw else -1
                    a00 += p0
                    a10 -= p0
                else:
                    da0 = 0
                    a00 = a10 = (a0 + a1) / 2

                da1 -= p1 * 2
                if da1 > EPSILON:
                    p1 *= 1 if cw else -1
                    a01 += p1
                    a11 -= p1
                else:
                    da1 = 0
                    a01 = a11 = (a0 + a1) / 2

            x01 = r1 * cos(a01)
            y01 = r1 * sin(a01)
            x10 = r0 * cos(a10)
            y10 = r0 * sin(a10)

            if rc > EPSILON:
                x11 = r1 * cos(a11)
                y11 = r1 * sin(a11)
                x00 = r0 * cos(a00)
                y00 = r0 * sin(a00)

                if da < pi:
                    if oc := intersect(x01, y01, x00, y00, x11, y11, x10, y10):
                        ax = x01 - oc[0]
                        ay = y01 - oc[1]
                        bx = x11 - oc[0]
                        by = y11 - oc[1]
                        kc = 1 / sin(
                            acos(
                                (ax * bx + ay * by)
                                / (sqrt(ax * ax + ay * ay) * sqrt(bx * bx + by * by))
                            )
                            / 2
                        )
                        lc = sqrt(oc[0] * oc[0] + oc[1] * oc[1])
                        rc0 = min(rc, (r0 - lc) / (kc - 1))
                        rc1 = min(rc, (r1 - lc) / (kc + 1))
                    else:
                        rc0 = rc1 = 0

            if da1 <= EPSILON:
                self._context.move_to(x01, y01)
            elif rc1 > EPSILON:
                t0 = CornerTangents(x00, y00, x01, y01, r1, rc1, cw)
                t1 = CornerTangents(x11, y11, x10, y10, r1, rc1, cw)
                self._context.move_to(t0.cx + t0.x01, t0.cy + t0.y01)

                if rc1 < rc:
                    self._context.arc(
                        t0.cx,
                        t0.cy,
                        rc1,
                        atan2(t0.y01, t0.x01),
                        atan2(t1.y01, t1.x01),
                        not cw,
                    )
                else:
                    self._context.arc(
                        t0.cx,
                        t0.cy,
                        rc1,
                        atan2(t0.y01, t0.x01),
                        atan2(t0.y11, t0.x11),
                        not cw,
                    )
                    self._context.arc(
                        0,
                        0,
                        r1,
                        atan2(t0.cy + t0.y11, t0.cx + t0.x11),
                        atan2(t1.cy + t1.y11, t1.cx + t1.x11),
                        not cw,
                    )
                    self._context.arc(
                        t1.cx,
                        t1.cy,
                        rc1,
                        atan2(t1.y11, t1.x11),
                        atan2(t1.y01, t1.x01),
                        not cw,
                    )
            else:
                self._context.move_to(x01, y01)
                self._context.arc(0, 0, r1, a01, a11, not cw)

            if r0 <= EPSILON or da0 <= EPSILON:
                self._context.line_to(x10, y10)
            elif rc0 > EPSILON:
                t0 = CornerTangents(x10, y10, x11, y11, r0, -rc0, cw)
                t1 = CornerTangents(x01, y01, x00, y00, r0, -rc0, cw)

                self._context.line_to(t0.cx + t0.x01, t0.cy + t0.y01)

                if rc0 < rc:
                    self._context.arc(
                        t0.cx,
                        t0.cy,
                        rc0,
                        atan2(t0.y01, t0.x01),
                        atan2(t1.y01, t1.x01),
                        not cw,
                    )
                else:
                    self._context.arc(
                        t0.cx,
                        t0.cy,
                        rc0,
                        atan2(t0.y01, t0.x01),
                        atan2(t0.y11, t0.x11),
                        not cw,
                    )
                    self._context.arc(
                        0,
                        0,
                        r0,
                        atan2(t0.cy + t0.y11, t0.cx + t0.x11),
                        atan2(t1.cy + t1.y11, t1.cx + t1.x11),
                        cw,
                    )
                    self._context.arc(
                        t1.cx,
                        t1.cy,
                        rc0,
                        atan2(t1.y11, t1.x11),
                        atan2(t1.y01, t1.x01),
                        not cw,
                    )
            else:
                self._context.arc(0, 0, r0, a10, a00, cw)

        self._context.close_path()
        if buffer is not None:
            self._context = None
            return str(buffer) or None

    def centroid(self, *args) -> tuple[float, float]:
        """
        Computes the midpoint [x, y] of the center line of
        the arc that would be generated by the given arguments.

        Parameters
        ----------
        args : Any
            Additional arguments

        Returns
        -------
        tuple[float, float]
            Midpoint coordinates
        """
        r = (self._inner_radius(*args) + self._outer_radius(*args)) * 0.5
        a = (self._start_angle(*args) + self._end_angle(*args)) * 0.5 - pi * 0.5
        return cos(a) * r, sin(a) * r

    def set_inner_radius(self, inner_radius: Callable[..., float] | float | int) -> Arc:
        """
        If radius is specified, sets the inner radius to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        inner_radius : Callable[..., float] | float | int
            Inner radius

        Returns
        -------
        Arc
            Itself
        """
        if callable(inner_radius):
            self._inner_radius = inner_radius
        else:
            self._inner_radius = constant(inner_radius)
        return self

    def set_outer_radius(self, outer_radius: Callable[..., float] | float | int) -> Arc:
        """
        If radius is specified, sets the outer radius to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        outer_radius : Callable[..., float] | float | int
            Outer radius

        Returns
        -------
        Arc
            Itself
        """
        if callable(outer_radius):
            self._outer_radius = outer_radius
        else:
            self._outer_radius = constant(outer_radius)
        return self

    def set_corner_radius(
        self, corner_radius: Callable[..., float] | float | int
    ) -> Arc:
        """
        If angle is specified, sets the corner angle to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        corner_radius : Callable[..., float] | float | int
            Corner radius

        Returns
        -------
        Arc
            Itself
        """
        if callable(corner_radius):
            self._corner_radius = corner_radius
        else:
            self._corner_radius = constant(corner_radius)
        return self

    def set_pad_radius(
        self, pad_radius: Callable[..., float] | float | int | None = None
    ) -> Arc:
        """
        If angle is specified, sets the pad radius to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        pad_radius : Callable[..., float] | float | int | None
            Pad radius

        Returns
        -------
        Arc
            Itself
        """
        if pad_radius is None:
            self._pad_radius = None
        elif callable(pad_radius):
            self._pad_radius = pad_radius
        else:
            self._pad_radius = constant(pad_radius)
        return self

    def set_start_angle(self, start_angle: Callable[..., float] | float | int) -> Arc:
        """
        If angle is specified, sets the start angle to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        start_angle : Callable[..., float] | float | int
            Start angle

        Returns
        -------
        Arc
            Itself
        """
        if callable(start_angle):
            self._start_angle = start_angle
        else:
            self._start_angle = constant(start_angle)
        return self

    def set_end_angle(self, end_angle: Callable[..., float] | float | int) -> Arc:
        """
        If angle is specified, sets the end angle to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        end_angle : Callable[..., float] | float | int
            End angle

        Returns
        -------
        Arc
            Itself
        """
        if callable(end_angle):
            self._end_angle = end_angle
        else:
            self._end_angle = constant(end_angle)
        return self

    def set_pad_angle(self, pad_angle: Callable[..., float] | float | int) -> Arc:
        """
        If angle is specified, sets the pad angle to the specified
        function or number and returns this arc generator.

        Parameters
        ----------
        pad_angle : Callable[..., float] | float | int
            Pad angle

        Returns
        -------
        Arc
            Itself
        """
        if callable(pad_angle):
            self._pad_angle = pad_angle
        else:
            self._pad_angle = constant(pad_angle)
        return self

    def set_context(self, context: Selection | None = None) -> Arc:
        """
        If context is specified, sets the context and
        returns this arc generator.

        Parameters
        ----------
        context : Selection | None
            Context

        Returns
        -------
        Arc
            Itself
        """
        self._context = context
        return self

    def get_inner_radius(self) -> Callable[..., float]:
        return self._inner_radius

    def get_outer_radius(self) -> Callable[..., float]:
        return self._outer_radius

    def get_corner_radius(self) -> Callable[..., float]:
        return self._corner_radius

    def get_pad_radius(self) -> Callable[..., float]:
        return self._pad_radius

    def get_pad_angle(self) -> Callable[..., float]:
        return self._pad_angle

    def get_context(self) -> Selection:
        return self._context



from ..types import Point2D


def area(polygon: list[Point2D]) -> float:
    """
    Returns the signed area of the specified polygon. If the vertices of the
    polygon are in counterclockwise order (assuming a coordinate system where
    the origin is in the top-left corner), the returned area is positive;
    otherwise it is negative, or zero.

    Parameters
    ----------
    polygon : list[Point2D]
        Polygon

    Returns
    -------
    float
        Signed area of the specified polygon

    Examples
    --------
    >>> d3.polygon_area([[1, 1], [1.5, 0], [2, 1]])
    -0.5
    """
    b = polygon[-1]
    area = 0
    for point in polygon:
        a = b
        b = point
        area += a[1] * b[0] - a[0] * b[1]

    return area * 0.5



from collections.abc import Callable

from ..color import hcl as color_hcl
from .color import color, hue


class HLCInterpolator:
    def __init__(self, func):
        self.func = func

    def __call__(self, start, end):
        start = color_hcl(start)
        end = color_hcl(end)
        h = self.func(start.h, end.h)
        c = color(start.c, end.c)
        l = color(start.l, end.l)
        opacity = color(start.opacity, end.opacity)

        def interpolate(t):
            start.h = h(t)
            start.c = c(t)
            start.l = l(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate


def interpolate_hcl(a: str, b: str) -> Callable[[float], str]:
    """
    Returns a CIELChab color space interpolator between the two colors a and b.
    The colors a and b need not be in CIELChab; they will be converted to CIELChab
    using d3.hcl. If either color's hue or chroma is NaN, the opposing color's channel
    value is used. The shortest path between hues is used. The return value of the
    interpolator is an RGB string.

    Parameters
    ----------
    a : str
        Color string a
    b : str
        Color string b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_hcl("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(245, 0, 134)'
    """
    return HLCInterpolator(hue)(a, b)


def interpolate_hcl_long(a: str, b: str) -> Callable[[float], str]:
    """
    Like :code:`interpolate_hcl`, but does not use the shortest path between hues.

    Parameters
    ----------
    a : str
        Color string a
    b : str
        Color string b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_hcl_long("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(0, 130, 64)'
    """
    return HLCInterpolator(color)(a, b)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "f7fcb9addd8e31a354",
    "ffffccc2e69978c679238443",
    "ffffccc2e69978c67931a354006837",
    "ffffccd9f0a3addd8e78c67931a354006837",
    "ffffccd9f0a3addd8e78c67941ab5d238443005a32",
    "ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443005a32",
    "ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443006837004529",
]
SCHEME_YLGN = list(map(colors, scheme))
interpolate_ylgn = ramp(SCHEME_YLGN)



from ..types import Point2D
from .common import SpatialTransform


class Compose:
    def __init__(self, a: SpatialTransform, b: SpatialTransform):
        self._a = a
        self._b = b
        self._invert = self._default_invert
        if hasattr(self._a, "invert") and hasattr(self._b, "invert"):
            self._invert = self._valid_invert

    def __call__(self, x: float, y: float) -> Point2D:
        a, b = self._a(x, y)
        return self._b(a, b)

    def invert(self, x: float, y: float) -> Point2D:
        return self._invert(x, y)

    def _valid_invert(self, x: float, y: float) -> Point2D:
        x = self._b.invert(x, y)
        return None if x is None else self._a.invert(x[0], x[1])

    def _default_invert(self, x: float, y: float):
        return

    def __str__(self) -> str:
        return f"Compose({self._a}, {self._b})"



from .basis import interpolate_basis
from .basis_closed import interpolate_basis_closed
from .cubehelix import interpolate_cubehelix, interpolate_cubehelix_long
from .date import interpolate_date
from .discrete import interpolate_discrete
from .hcl import interpolate_hcl, interpolate_hcl_long
from .hsl import interpolate_hsl, interpolate_hsl_long
from .hue import interpolate_hue
from .lab import interpolate_lab
from .number import interpolate_number
from .number_array import interpolate_number_array
from .object_to_value import interpolate, interpolate_array, interpolate_object
from .piecewise import piecewise
from .quantize import quantize
from .rgb import interpolate_rgb, interpolate_rgb_basis, interpolate_rgb_basis_closed
from .round import interpolate_round
from .string import interpolate_string
from .zoom import interpolate_zoom

__all__ = [
    "interpolate",
    "interpolate_array",
    "interpolate_basis",
    "interpolate_basis_closed",
    "interpolate_cubehelix",
    "interpolate_cubehelix_long",
    "interpolate_date",
    "interpolate_discrete",
    "interpolate_hcl",
    "interpolate_hcl_long",
    "interpolate_hsl",
    "interpolate_hsl_long",
    "interpolate_hue",
    "interpolate_lab",
    "interpolate_number",
    "interpolate_number_array",
    "interpolate_object",
    "interpolate_rgb",
    "interpolate_rgb_basis",
    "interpolate_rgb_basis_closed",
    "interpolate_round",
    "interpolate_string",
    "interpolate_zoom",
    "piecewise",
    "quantize",
]



import math
from statistics import stdev


def threshold_scott(values: list[float | None], mini: float, maxi: float) -> int:
    """
    Returns the number of bins according to Scott's normal reference rule; the
    input values must be numbers.

    Parameters
    ----------
    values : list[float | None]
        Input values
    mini : float
        Minimum value
    maxi : float
        Maximum value

    Returns
    -------
    int
       Number of bins according to Scott's normal
    """
    values = [v for v in values if v is not None]
    c = len(values)
    if len(values) <= 1:
        d = 0
    else:
        d = stdev(values)
    return math.ceil((maxi - mini) * (c ** (1 / 3)) / (3.49 * d)) if c and d else 1



from __future__ import annotations


OVERSHOOT = 1.70158


class BackIn:
    """
    Anticipatory easing like a dancer bending her knees before jumping off the
    floor.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_back_in(0.2)
    -0.04645056000000002
    >>> d3.ease_back_in.overshoot(1.5)(0.2)
    -0.040000000000000015
    """

    def __init__(self, overshoot: float):
        self._overshoot = overshoot

    def __call__(self, t: float) -> float:
        """
        Anticipatory easing like a dancer bending her knees before jumping off
        the floor.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_back_in(0.2)
        -0.04645056000000002
        >>> d3.ease_back_in.overshoot(1.5)(0.2)
        -0.040000000000000015
        """
        return t * t * (self._overshoot * (t - 1) + t)

    def overshoot(self, overshoot: float) -> BackIn:
        """
        Returns a new easing function with :code:`overshoot` value as
        parameter.

        Parameters
        ----------
        overshoot : float
            Overshoot value

        Returns
        -------
        BackIn
            New easing function
        """
        return BackIn(overshoot)


ease_back_in = BackIn(OVERSHOOT)


class BackOut:
    """
    Reverse anticipatory easing; equivalent to :code:`1 - d3.ease_back_in(1
    - t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_back_out(0.2)
    0.7058022399999999
    >>> d3.ease_back_out.overshoot(1.5)(0.2)
    0.6799999999999999
    """

    def __init__(self, overshoot: float):
        self._overshoot = overshoot

    def __call__(self, t: float) -> float:
        """
        Reverse anticipatory easing; equivalent to :code:`1 - d3.ease_back_in(1
        - t)`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_back_out(0.2)
        0.7058022399999999
        >>> d3.ease_back_out.overshoot(1.5)(0.2)
        0.6799999999999999
        """
        t -= 1
        return t * t * ((t + 1) * self._overshoot + t) + 1

    def overshoot(self, overshoot: float) -> BackOut:
        """
        Returns a new easing function with :code:`overshoot` value as
        parameter.

        Parameters
        ----------
        overshoot : float
            Overshoot value

        Returns
        -------
        BackOut
            New easing function
        """
        return BackOut(overshoot)


ease_back_out = BackOut(OVERSHOOT)


class BackInOut:
    """
    Symmetric anticipatory easing; scales :func:`d3.ease_back_in
    <detroit.ease_back_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_back_out <detroit.ease_back_out>` :math:`\\forall t \\in
    [0.5, 1]`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_back_in_out(0.2)
    -0.04967584000000001
    >>> d3.ease_back_in_out.overshoot(1.5)(0.2)
    -0.04000000000000001
    """

    def __init__(self, overshoot: float):
        self._overshoot = overshoot

    def __call__(self, t: float) -> float:
        """
        Symmetric anticipatory easing; scales :func:`d3.ease_back_in
        <detroit.ease_back_in>` :math:`\\forall t \\in [0, 0.5]` and
        :func:`d3.ease_back_out <detroit.ease_back_out>` :math:`\\forall t \\in
        [0.5, 1]`.

        Parameters
        ----------
        t : float
            :math:`t \\in [0, 1]`

        Returns
        -------
        float
            Output value of the easing function

        Examples
        --------
        >>> d3.ease_back_in_out(0.2)
        -0.04967584000000001
        >>> d3.ease_back_in_out.overshoot(1.5)(0.2)
        -0.04000000000000001
        """
        t *= 2
        if t < 1:
            return t * t * ((self._overshoot + 1) * t - self._overshoot) * 0.5
        else:
            t -= 2
            return (t * t * ((self._overshoot + 1) * t + self._overshoot) + 2) * 0.5

    def overshoot(self, overshoot: float) -> BackInOut:
        """
        Returns a new easing function with :code:`overshoot` value as
        parameter.

        Parameters
        ----------
        overshoot : float
            Overshoot value

        Returns
        -------
        BackInOut
            New easing function
        """
        return BackInOut(overshoot)


ease_back_in_out = BackInOut(OVERSHOOT)



import math

e10 = math.sqrt(50)
e5 = math.sqrt(10)
e2 = math.sqrt(2)


def tick_spec(start: float, stop: float, count: int) -> tuple[int, int, int]:
    step = (stop - start) / max(0, count)
    power = math.floor(math.log10(step))
    error = step / (10**power)
    factor = 10 if error >= e10 else 5 if error >= e5 else 2 if error >= e2 else 1
    if power < 0:
        inc = (10**-power) / factor
        i1 = round(start * inc)
        i2 = round(stop * inc)
        if i1 / inc < start:
            i1 += 1
        if i2 / inc > stop:
            i2 -= 1
        inc = -inc
    else:
        inc = (10**power) * factor
        i1 = round(start / inc)
        i2 = round(stop / inc)
        if i1 * inc < start:
            i1 += 1
        if i2 * inc > stop:
            i2 -= 1
    if i2 < i1 and 0.5 <= count < 2:
        return tick_spec(start, stop, count * 2)
    return [i1, i2, inc]


def ticks(
    start: int | float | str, stop: int | float | str, count: int | float
) -> list[float]:
    """
    Returns an array of approximately count + 1 uniformly-spaced,
    nicely-rounded values between start and stop (inclusive).

    Parameters
    ----------
    start : int | float | str
        Start value which can be converted as float
    stop : int | float | str
        Stop value which can be converted as float
    count : int | float
        Count value

    Returns
    -------
    list[float]
        Array of nicely-rounded values

    Examples
    --------

    >>> d3.ticks(0, 1, 10)
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    >>> d3.ticks("1.2", "2", 10)
    [1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0]
    """
    stop = float(stop)
    start = float(start)
    count = float(count)
    if count <= 0:
        return []
    if start == stop:
        return [start]
    reverse = stop < start
    i1, i2, inc = (
        tick_spec(stop, start, count) if reverse else tick_spec(start, stop, count)
    )
    if i2 < i1:
        return []
    n = i2 - i1 + 1
    tick_list = [0] * n
    if reverse:
        if inc < 0:
            for i in range(n):
                tick_list[i] = (i2 - i) / -inc
        else:
            for i in range(n):
                tick_list[i] = (i2 - i) * inc
    else:
        if inc < 0:
            for i in range(n):
                tick_list[i] = (i1 + i) / -inc
        else:
            for i in range(n):
                tick_list[i] = (i1 + i) * inc
    return tick_list


def tick_increment(
    start: int | float | str, stop: int | float | str, count: int | float
) -> int:
    """
    Returns the negative inverse tick step instead.

    Parameters
    ----------
    start : int | float | str
        Start value which can be converted as float
    stop : int | float | str
        Stop value which can be converted as float
    count : int | float
        Count value

    Returns
    -------
    int
        Increment value

    Examples
    --------

    >>> d3.tick_increment(0, 1, 10)
    -10.0
    >>> d3.tick_increment(0, 1, 1)
    1
    >>> d3.tick_increment(0, 1, 100)
    -100.0
    >>> d3.tick_increment(0, 20, 10)
    2
    """
    stop = float(stop)
    start = float(start)
    count = float(count)
    return tick_spec(start, stop, count)[2]


def tick_step(
    start: int | float | str, stop: int | float | str, count: int | float
) -> float:
    """
    Returns the difference between adjacent tick value.

    Parameters
    ----------
    start : int | float | str
        Start value which can be converted as float
    stop : int | float | str
        Stop value which can be converted as float
    count : int | float
        Count value

    Returns
    -------
    int
        Increment value

    Examples
    --------

    >>> d3.tick_step(0, 10, 1)
    10
    >>> d3.tick_step(0, 1, 10)
    0.1
    """
    stop = float(stop)
    start = float(start)
    count = float(count)
    if stop == start:
        return 1
    reverse = stop < start
    inc = (
        tick_increment(stop, start, count)
        if reverse
        else tick_increment(start, stop, count)
    )
    return (1 if not reverse else -1) * (1 / -inc if inc < 0 else inc)



from .band import scale_band, scale_point
from .diverging import (
    scale_diverging,
    scale_diverging_log,
    scale_diverging_pow,
    scale_diverging_sqrt,
    scale_diverging_symlog,
)
from .identity import scale_identity
from .linear import scale_linear
from .log import scale_log
from .ordinal import scale_ordinal
from .pow import scale_pow, scale_sqrt
from .quantile import scale_quantile
from .quantize import scale_quantize
from .radial import scale_radial
from .sequential import (
    scale_sequential,
    scale_sequential_log,
    scale_sequential_pow,
    scale_sequential_sqrt,
    scale_sequential_symlog,
)
from .sequential_quantile import scale_sequential_quantile
from .symlog import scale_symlog
from .threshold import scale_threshold
from .tick_format import tick_format
from .time import scale_time

__all__ = [
    "scale_band",
    "scale_diverging",
    "scale_diverging",
    "scale_diverging_log",
    "scale_diverging_log",
    "scale_diverging_pow",
    "scale_diverging_pow",
    "scale_diverging_sqrt",
    "scale_diverging_sqrt",
    "scale_diverging_symlog",
    "scale_diverging_symlog",
    "scale_identity",
    "scale_linear",
    "scale_log",
    "scale_ordinal",
    "scale_point",
    "scale_pow",
    "scale_quantile",
    "scale_quantize",
    "scale_radial",
    "scale_sequential",
    "scale_sequential_log",
    "scale_sequential_pow",
    "scale_sequential_quantile",
    "scale_sequential_sqrt",
    "scale_sequential_symlog",
    "scale_sqrt",
    "scale_symlog",
    "scale_threshold",
    "scale_time",
    "tick_format",
]



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e5f5f999d8c92ca25f",
    "edf8fbb2e2e266c2a4238b45",
    "edf8fbb2e2e266c2a42ca25f006d2c",
    "edf8fbccece699d8c966c2a42ca25f006d2c",
    "edf8fbccece699d8c966c2a441ae76238b45005824",
    "f7fcfde5f5f9ccece699d8c966c2a441ae76238b45005824",
    "f7fcfde5f5f9ccece699d8c966c2a441ae76238b45006d2c00441b",
]
SCHEME_BUGN = list(map(colors, scheme))
interpolate_bugn = ramp(SCHEME_BUGN)



from __future__ import annotations

import math
from typing import overload

from .color import RGB, Color, rgb_convert

K = 18
XN = 0.96422
YN = 1
ZN = 0.82521
T0 = 4 / 29
T1 = 6 / 29
T2 = 3 * T1 * T1
T3 = T1 * T1 * T1


def lab_convert(obj):
    if isinstance(obj, LAB):
        return LAB(obj.l, obj.a, obj.b, obj.opacity)
    if isinstance(obj, HCL):
        return hcl2lab(obj)
    if not isinstance(obj, RGB):
        obj = rgb_convert(obj)
    r = rgb2lrgb(obj.r)
    g = rgb2lrgb(obj.g)
    b = rgb2lrgb(obj.b)
    y = xyz2lab((0.2225045 * r + 0.7168786 * g + 0.0606169 * b) / YN)
    if r == g == b:
        x = z = y
    else:
        x = xyz2lab((0.4360747 * r + 0.3850649 * g + 0.1430804 * b) / XN)
        z = xyz2lab((0.0139322 * r + 0.0971045 * g + 0.7141733 * b) / ZN)
    return LAB(116 * y - 16, 500 * (x - y), 200 * (y - z), obj.opacity)


def xyz2lab(t):
    return t ** (1 / 3) if t > T3 else t / T2 + T0


def lab2xyz(t):
    return t**3 if t > T1 else T2 * (t - T0)


def lrgb2rgb(x):
    return 255 * (12.92 * x if x <= 0.0031308 else 1.055 * x ** (1 / 2.4) - 0.055)


def rgb2lrgb(x):
    x /= 255
    return x / 12.92 if x <= 0.04045 else ((x + 0.055) / 1.055) ** 2.4


def hcl_convert(obj):
    if isinstance(obj, HCL):
        return HCL(obj.h, obj.c, obj.l, obj.opacity)
    if not isinstance(obj, LAB):
        obj = lab_convert(obj)
    if obj.a == 0 and obj.b == 0:
        return HCL(
            float("nan"), 0 if 0 < obj.l < 100 else float("nan"), obj.l, obj.opacity
        )
    h = math.degrees(math.atan2(obj.b, obj.a))
    return HCL(
        h + 360 if h < 0 else h, (obj.a**2 + obj.b**2) ** 0.5, obj.l, obj.opacity
    )


def hcl2lab(obj):
    if math.isnan(obj.h):
        return LAB(obj.l, 0, 0, obj.opacity)
    h = math.radians(obj.h)
    return LAB(obj.l, math.cos(h) * obj.c, math.sin(h) * obj.c, obj.opacity)


class LAB(Color):
    """
    LAB color format

    Parameters
    ----------
    l : float
        L* channel value
    a : float
        a* channel value
    b : float
        b* channel value
    opacity : float
        Opacity value
    """

    def __init__(self, l: float, a: float, b: float, opacity: float = 1.0):
        self.l = float(l)
        self.a = float(a)
        self.b = float(b)
        self.opacity = float(opacity)

    def brighter(self, k: float | None = None) -> LAB:
        """
        Returns a brighter copy of this color.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        LAB
            Brighter LAB
        """
        return LAB(self.l + K * (1 if k is None else k), self.a, self.b, self.opacity)

    def darker(self, k: float | None = None) -> LAB:
        """
        Returns a darker copy of this color.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        LAB
            Brighter LAB
        """
        return LAB(self.l - K * (1 if k is None else k), self.a, self.b, self.opacity)

    def rgb(self) -> RGB:
        """
        Returns the RGB equivalent of this color

        Returns
        -------
        RGB
            RGB color format
        """
        y = (self.l + 16) / 116
        x = y if math.isnan(self.a) else y + self.a / 500
        z = y if math.isnan(self.b) else y - self.b / 200
        x = XN * lab2xyz(x)
        y = YN * lab2xyz(y)
        z = ZN * lab2xyz(z)
        return RGB(
            lrgb2rgb(3.1338561 * x - 1.6168667 * y - 0.4906146 * z),
            lrgb2rgb(-0.9787684 * x + 1.9161415 * y + 0.0334540 * z),
            lrgb2rgb(0.0719453 * x - 0.2289914 * y + 1.4052427 * z),
            self.opacity,
        )

    def __repr__(self) -> str:
        return f"LAB(l={self.l}, a={self.a}, b={self.b}, opacity={self.opacity})"


class HCL(Color):
    """
    HCL color format

    Parameters
    ----------
    h : float
        Hue channel value
    c : float
        Chroma channel value
    l : float
        Luminance channel value
    opacity : float
        Opacity value
    """

    def __init__(self, h: float, c: float, l: float, opacity: float = 1.0):
        self.h = float(h)
        self.c = float(c)
        self.l = float(l)
        self.opacity = float(opacity)

    def brighter(self, k: float | None = None) -> HCL:
        """
        Returns a brighter copy of this color.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        HCL
            Brighter HCL
        """
        return HCL(self.h, self.c, self.l + K * (1 if k is None else k), self.opacity)

    def darker(self, k: float | None = None) -> HCL:
        """
        Returns a darker copy of this color.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        HCL
            Brighter HCL
        """
        return HCL(self.h, self.c, self.l - K * (1 if k is None else k), self.opacity)

    def rgb(self) -> RGB:
        """
        Returns the RGB equivalent of this color

        Returns
        -------
        RGB
            RGB color format
        """
        return hcl2lab(self).rgb()

    def __repr__(self) -> str:
        return f"HCL(h={self.h}, c={self.c}, l={self.l}, opacity={self.opacity})"


@overload
def lab(specifier: str) -> LAB: ...


@overload
def lab(l: float, a: float, b: float) -> LAB: ...


@overload
def lab(l: float, a: float, b: float, opacity: float) -> LAB: ...


def lab(*args):
    """
    Builds a new LAB color.

    Parameters
    ----------
    specifier : str
        String which represents a color
    l : float
        L* channel value between 0 and 100
    a : float
        a* channel value between -128 and 127
    b : float
        b* channel value between -128 and 127
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    LAB
        LAB object

    Examples
    --------

    >>> d3.lab("#2e65ffff")
    LAB(l=46.97291122702186, a=27.037050613700043, b=-83.1712687492733, opacity=1.0)
    >>> d3.lab(78.25, -17.71, 76.47, 0.8)
    LAB(l=78.25, a=-17.71, b=76.47, opacity=0.8)
    """
    if len(args) == 1:
        return lab_convert(args[0])
    elif len(args) == 3:
        l, a, b = args
        opacity = 1
        return LAB(l, a, b, opacity)
    elif len(args) == 4:
        l, a, b, opacity = args
        return LAB(l, a, b, opacity)


@overload
def gray(l: float) -> LAB: ...


@overload
def gray(l: float, opacity: float) -> gray: ...


def gray(*args):
    """
    Builds a new LAB color on gray gradient.

    Parameters
    ----------
    l : float
        L* channel value between 0 and 100
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    LAB
        LAB object

    Examples
    --------

    >>> d3.gray(78.25, 0.8)
    LAB(l=78.25, a=0.0, b=0.0, opacity=0.8)
    """
    if len(args) == 1:
        l = args[0]
        opacity = 1
        return LAB(l, 0, 0, opacity)
    elif len(args) == 2:
        l, opacity = args
        return LAB(l, 0, 0, opacity)


@overload
def lch(specifier: str) -> HCL: ...


@overload
def lch(l: float, c: float, h: float) -> HCL: ...


@overload
def lch(l: float, c: float, h: float, opacity: float) -> HCL: ...


def lch(*args):
    """
    Builds a new HCL color from LCH format.

    Parameters
    ----------
    specifier : str
        String which represents a color
    l : float
        Luminance channel value
    c : float
        Chroma channel value
    h : float
        Hue channel value
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    HCL
        HCL object

    Examples
    --------

    >>> d3.lch("#2e65ffff")
    HCL(h=288.00814189783785, c=87.45548611294561, l=46.97291122702186, opacity=1.0)
    >>> d3.lch(43.14, 83.82, 60, 0.8)
    HCL(h=60.0, c=83.82, l=43.14, opacity=0.8)
    """
    if len(args) == 1:
        return hcl_convert(args[0])
    elif len(args) == 3:
        l, c, h = args
        opacity = 1
        return HCL(h, c, l, opacity)
    elif len(args) == 4:
        l, c, h, opacity = args
        return HCL(h, c, l, opacity)


@overload
def hcl(specifier: str) -> HCL: ...


@overload
def hcl(h: float, c: float, l: float) -> HCL: ...


@overload
def hcl(h: float, c: float, l: float, opacity: float) -> HCL: ...


def hcl(*args):
    """
    Builds a new HCL color.

    Parameters
    ----------
    specifier : str
        String which represents a color
    h : float
        Hue channel value
    c : float
        Chroma channel value
    l : float
        Luminance channel value
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    HCL
        HCL object

    Examples
    --------

    >>> d3.hcl("#2e65ffff")
    HCL(h=288.00814189783785, c=87.45548611294561, l=46.97291122702186, opacity=1.0)
    >>> d3.hcl(60, 83.82, 43.14, 0.8)
    HCL(h=60.0, c=83.82, l=43.14, opacity=0.8)
    """
    if len(args) == 1:
        return hcl_convert(args[0])
    elif len(args) == 3:
        h, c, l = args
        opacity = 1
        return HCL(h, c, l, opacity)
    elif len(args) == 4:
        h, c, l, opacity = args
        return HCL(h, c, l, opacity)



from ..types import Point2D


def centroid(polygon: list[Point2D]) -> Point2D:
    """
    Returns the centroid of the specified polygon.

    Parameters
    ----------
    polygon : list[Point2D]
        Polygon

    Returns
    -------
    Point2D
        Centroid point

    Examples
    --------
    >>> d3.polygon_centroid([[1, 1], [1.5, 0], [2, 1]])
    [1.5, 0.6666666666666666]
    """
    x = 0
    y = 0
    b = polygon[-1]
    k = 0

    for point in polygon:
        a = b
        b = point
        c = a[0] * b[1] - b[0] * a[1]
        k += c
        x += (a[0] + b[0]) * c
        y += (a[1] + b[1]) * c

    k *= 3
    return [x / k, y / k]



import math
from collections.abc import Callable

from ..array import tick_step
from ..format import (
    format_prefix,
    format_specifier,
    locale_format,
    precision_fixed,
    precision_prefix,
    precision_round,
)
from ..types import T


def tick_format(
    start: T, stop: T, count: int, specifier: str | None = None
) -> Callable[[T], str]:
    """
    Returns a number format function suitable for displaying a tick value,
    automatically computing the appropriate precision based on the fixed
    interval between tick values, as determined by d3.tickStep.

    Parameters
    ----------
    start : T
        Start value
    stop : T
        Stop value
    count : int
        Count value
    specifier : str | None
        Specifier allows a custom format where the precision of the format
        is automatically set by the scale as appropriate for the tick interval.
    """
    step = tick_step(start, stop, count)
    precision = None
    specifier = format_specifier(specifier if specifier is not None else ",f")

    if specifier.type == "s":
        value = max(abs(start), abs(stop))
        if specifier.precision is None and not math.isnan(
            precision := precision_prefix(step, value)
        ):
            specifier.precision = precision
        return format_prefix(specifier, value)
    elif specifier.type in ("", "e", "g", "p", "r"):
        if specifier.precision is None and not math.isnan(
            precision := precision_round(step, max(abs(start), abs(stop)))
        ):
            specifier.precision = precision - (specifier.type == "e")
    elif specifier.type in ("f", "%"):
        if specifier.precision is None and not math.isnan(
            precision := precision_fixed(step)
        ):
            specifier.precision = precision - (specifier.type == "%") * 2

    return locale_format(specifier)



from .area import area as polygon_area
from .centroid import centroid as polygon_centroid
from .contains import contains as polygon_contains
from .hull import hull as polygon_hull
from .length import length as polygon_length

__all__ = [
    "polygon_area",
    "polygon_centroid",
    "polygon_contains",
    "polygon_hull",
    "polygon_length",
]



import math

from ...selection import Selection
from .basis import BezierTrait
from .common import Curve


class BasisClosedCurve(Curve, BezierTrait):
    def __init__(self, context):
        self._context = context
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._x2 = math.nan
        self._y2 = math.nan
        self._x3 = math.nan
        self._y3 = math.nan
        self._x4 = math.nan
        self._y4 = math.nan

    def area_start(self):
        return None

    def area_end(self):
        return None

    def line_start(self):
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._x2 = math.nan
        self._y2 = math.nan
        self._x3 = math.nan
        self._y3 = math.nan
        self._x4 = math.nan
        self._y4 = math.nan
        self._point = 0

    def line_end(self):
        if self._point == 1:
            self._context.move_to(self._x2, self._y2)
            self._context.close_path()
        elif self._point == 2:
            self._context.move_to(
                (self._x2 + 2 * self._x3) / 3,
                (self._y2 + 2 * self._y3) / 3,
            )
            self._context.line_to(
                (self._x3 + 2 * self._x2) / 3,
                (self._y3 + 2 * self._y2) / 3,
            )
            self._context.close_path()
        elif self._point == 3:
            self.point(self._x2, self._y2)
            self.point(self._x3, self._y3)
            self.point(self._x4, self._y4)

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            self._x2 = x
            self._y2 = y
        elif self._point == 1:
            self._point = 2
            self._x3 = x
            self._y3 = y
        elif self._point == 2:
            self._point = 3
            self._x4 = x
            self._y4 = y
            self._context.move_to(
                (self._x0 + 4 * self._x1 + x) / 6,
                (self._y0 + 4 * self._y1 + y) / 6,
            )
        else:
            self._bezier_curve_to(x, y)
        self._x0 = self._x1
        self._x1 = x
        self._y0 = self._y1
        self._y1 = y


def curve_basis_closed(context: Selection) -> Curve:
    """
    Produces a closed cubic basis spline using the specified control points.
    When a line segment ends, the first three control points are repeated,
    producing a closed loop with :math:`C^2` continuity.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_basis_closed)
    >>> line(points)
    'M6,7.333C7.333,8.667,8.667,9.333,9.667,9.667C10.667,10,11.333,10,12,9C12.667,8,13.333,6,14.667,5C16,4,18,4,19.667,4.667C21.333,5.333,22.667,6.667,24.167,7C25.667,7.333,27.333,6.667,28.667,6C30,5.333,31,4.667,32,4.500C33,4.333,34,4.667,35,4.333C36,4,37,3,31.500,2.500C26,2,14,2,8.667,3C3.333,4,4.667,6,6,7.333'

    **Result**

    .. image:: ../../figures/light_curve_basis_closed.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_basis_closed.svg
       :align: center
       :class: only-dark
    """
    return BasisClosedCurve(context)



from math import atan2, cos, fsum, pi, radians, sin

from ..array import argpass
from ..types import GeoJSON
from .common import PolygonStream
from .stream import geo_stream

QUARTER_PI = pi * 0.25
TAU = 2 * pi


@argpass
def noop():
    return


class AreaStream(PolygonStream):
    def __init__(self):
        self._point = noop
        self._line_start = noop
        self._line_end = noop

        self._area_ring_sum = []
        self._area_sum = []

        self._lambda00 = None
        self._phi00 = None
        self._lambda0 = None
        self._cos_phi0 = None
        self._sin_phi0 = None

    def point(self, x: float, y: float):
        return self._point(x, y)

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._area_ring_sum = []
        self._line_start = self._area_ring_start
        self._line_end = self._area_ring_end

    def polygon_end(self):
        area_ring = fsum(self._area_ring_sum)
        area_ring = TAU + area_ring if area_ring < 0 else area_ring
        self._area_sum.append(area_ring)
        self._point = noop
        self._line_start = noop
        self._line_end = noop

    def sphere(self):
        self._area_sum.append(TAU)

    def _area_ring_start(self):
        self._point = self._area_point_first

    def _area_ring_end(self):
        self._area_point(self._lambda00, self._phi00)

    def _area_point_first(self, lambda_: float, phi: float):
        self._point = self._area_point
        self._lambda00 = lambda_
        self._phi00 = phi
        lambda_ = radians(lambda_)
        phi = radians(phi)
        self._lambda0 = lambda_
        phi = phi * 0.5 + QUARTER_PI
        self._cos_phi0 = cos(phi)
        self._sin_phi0 = sin(phi)

    def _area_point(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        phi = phi * 0.5 + QUARTER_PI

        d_lambda = lambda_ - self._lambda0
        sd_lambda = 1 if d_lambda >= 0 else -1
        ad_lambda = sd_lambda * d_lambda
        cos_phi = cos(phi)
        sin_phi = sin(phi)
        k = self._sin_phi0 * sin_phi
        u = self._cos_phi0 * cos_phi + k * cos(ad_lambda)
        v = k * sd_lambda * sin(ad_lambda)

        self._area_ring_sum.append(atan2(v, u))
        self._lambda0 = lambda_
        self._cos_phi0 = cos_phi
        self._sin_phi0 = sin_phi

    def result(self) -> float:
        return fsum(self._area_sum) * 2


def geo_area(obj: GeoJSON) -> float:
    """
    Returns the spherical area of the specified GeoJSON object in steradians.
    This is the spherical equivalent of :func:`GeoPath.area
    <detroit.geo.path.path.GeoPath.area>`.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object

    Returns
    -------
    float
        Spherical area
    """
    stream = AreaStream()
    geo_stream(obj, stream)
    return stream.result()



from __future__ import annotations

from collections.abc import Callable
from operator import itemgetter
from typing import Any, Generic

from ..array import argpass
from ..selection import Selection
from ..types import Accessor, Number, T
from .constant import constant
from .curves import Curve, curve_bump_radial, curve_bump_x, curve_bump_y
from .path import WithPath
from .point import x as point_x
from .point import y as point_y

link_source = itemgetter("source")
link_target = itemgetter("target")


class Link(Generic[T], WithPath):
    """
    The link shape generates a smooth cubic Bézier curve from a source point to
    a target point. The tangents of the curve at the start and end are either
    vertical or horizontal.

    Parameters
    ----------
    curve : Callable[[Selection], Curve]
        Curve factory function

    Returns
    -------
    Link
        New link generator
    """

    def __init__(self, curve: Callable[[Selection], Curve]):
        super().__init__()
        self._source = argpass(link_source)
        self._target = argpass(link_target)
        self._x = argpass(point_x)
        self._y = argpass(point_y)
        self._context = None
        self._output = None
        self._curve = curve

    def __call__(self, *args: Any) -> str | None:
        """

        Parameters
        ----------
        args : Any
            Extra arguments passed through :code:`source` function,
            :code:`target` function, :code:`x` function and :code:`y` function.

        Returns
        -------
        str | None
            Generated link if the link is not associated to a context
        """
        buffer = None
        args = list(args)
        s = self._source(*args)
        t = self._target(*args)
        if self._context is None:
            buffer = self._path()
            self._output = self._curve(buffer)
        self._output.line_start()
        args[0] = s
        x = self._x(*args)
        y = self._y(*args)
        self._output.point(x, y)
        args[0] = t
        x = self._x(*args)
        y = self._y(*args)
        self._output.point(x, y)
        self._output.line_end()
        if buffer:
            self._output = None
            return str(buffer) or None

    def set_source(self, source: Accessor[T, float]) -> Link:
        """
        Sets the source function

        Parameters
        ----------
        source : Accessor[T, float]
            Source accessor function

        Returns
        -------
        Link
            Itself
        """
        self._source = argpass(source)
        return self

    def set_target(self, target: Accessor[T, float]) -> Link:
        """
        Sets the target function

        Parameters
        ----------
        target : Accessor[T, float]
            Target accessor function

        Returns
        -------
        Link
            Itself
        """
        self._target = argpass(target)
        return self

    def x(self, x: Accessor[T, float] | Number) -> Link:
        """
        Sets x accessor function

        Parameters
        ----------
        x : Accessor[T, float] | Number
            x accessor function

        Returns
        -------
        Link
            Itself
        """
        if callable(x):
            self._x = x
        else:
            self._x = constant(x)
        self._x = argpass(self._x)
        return self

    def y(self, y: Accessor[T, float] | Number) -> Link:
        """
        Sets y accessor function

        Parameters
        ----------
        y : Accessor[T, float] | Number
            y accessor function

        Returns
        -------
        Link
            Itself
        """
        if callable(y):
            self._y = y
        else:
            self._y = constant(y)
        self._y = argpass(self._y)
        return self

    def set_context(self, context: Selection | None = None) -> Link:
        """
        Sets the context.

        Parameters
        ----------
        context : Selection | None
            Selection

        Returns
        -------
        Link
            Itself
        """
        if context is None:
            self._context = None
            self._output = None
        else:
            self._context = context
            self._output = self._curve(self._context)
        return self

    def get_source(self) -> Accessor[T, float]:
        return self._source

    def get_target(self) -> Accessor[T, float]:
        return self._target

    def get_x(self) -> Accessor[T, float]:
        return self._x

    def get_y(self) -> Accessor[T, float]:
        return self._y

    def get_context(self) -> Selection:
        return self._context


def link_horizontal() -> Link:
    """
    Shorthand for link with :func:`d3.curve_bump_x <curve_bump_x>`; suitable
    for visualizing links in a tree diagram rooted on the left edge of the
    display.

    Returns
    -------
    Link
        Horizontal link generator
    """
    return Link(curve_bump_x)


def link_vertical() -> Link:
    """
    Shorthand for link :func:`d3.curve_bump_y <curve_bump_y>`; suitable for
    visualizing links in a tree diagram rooted on the top edge of the display.

    Returns
    -------
    Link
        Vertical link generator
    """
    return Link(curve_bump_y)


def link_radial() -> Link:
    """
    Returns a new link generator with radial tangents.

    Returns
    -------
    Link
        Radial link generator
    """
    link = Link(curve_bump_radial)
    link.angle = link.x
    link.radius = link.y
    link.get_angle = link.get_x
    link.get_radius = link.get_y
    return link



import re

# [[fill]align][sign][symbol][0][width][,][.precision][~][type]
FORMAT_RE = re.compile(
    r"^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$", re.I
)


class FormatSpecifier:
    def __init__(self, specifier):
        self.fill = " " if specifier.get("fill") is None else str(specifier["fill"])
        self.align = ">" if specifier.get("align") is None else str(specifier["align"])
        self.sign = "-" if specifier.get("sign") is None else str(specifier["sign"])
        self.symbol = (
            "" if specifier.get("symbol") is None else str(specifier["symbol"])
        )
        self.zero = bool(specifier.get("zero"))
        self.width = None if specifier.get("width") is None else int(specifier["width"])
        self.comma = bool(specifier.get("comma"))
        self.precision = (
            None if specifier.get("precision") is None else int(specifier["precision"])
        )
        self.trim = bool(specifier.get("trim"))
        self.type = "" if specifier.get("type") is None else str(specifier["type"])

    def __str__(self):
        parts = [
            self.fill,
            self.align,
            self.sign,
            self.symbol,
            "0" if self.zero else "",
            str(self.width) if self.width is not None else "",
            "," if self.comma else "",
            f".{self.precision}" if self.precision is not None else "",
            "~" if self.trim else "",
            self.type,
        ]
        return "".join(parts)


def format_specifier(specifier: str) -> FormatSpecifier:
    match = FORMAT_RE.match(specifier)
    if not match:
        raise ValueError(f"invalid format: {specifier}")

    groups = match.groups()
    return FormatSpecifier(
        {
            "fill": groups[0],
            "align": groups[1],
            "sign": groups[2],
            "symbol": groups[3],
            "zero": groups[4],
            "width": groups[5],
            "comma": groups[6],
            "precision": groups[7][1:] if groups[7] else None,
            "trim": groups[8],
            "type": groups[9],
        }
    )



from __future__ import annotations

from collections.abc import Callable
from functools import cmp_to_key

from ..common import LineStream, Point2D
from ..point_equal import EPSILON, point_equal


class Intersection:
    __slots__ = ("x", "z", "o", "e", "v", "n", "p")

    def __init__(
        self, point: Point2D, points: list[Point2D], other: Intersection, entry: bool
    ):
        self.x = point
        self.z = points
        self.o = other
        self.e = entry
        self.v = False
        self.n = None
        self.p = None


def clip_rejoin(
    segments: list[list[tuple[float, float, float | None]]],
    compare_intersection: Callable[[Intersection, Intersection], float],
    start_inside: int,
    interpolate: Callable[[float | None, float | None, float, LineStream], None],
    stream: LineStream,
):
    subject = []
    clip = []

    if segments:
        for segment in segments:
            n = len(segment) - 1
            if n <= 0:
                continue
            p0 = segment[0]
            p1 = segment[n]

            if point_equal(p0, p1):
                if not p0[2] and not p1[2]:
                    stream.line_start()
                    for i in range(n):
                        p0 = segment[i]
                        stream.point(p0[0], p0[1])
                    stream.line_end()
                    continue
                p1[0] += 2 * EPSILON

            x = Intersection(p0, segment, None, True)
            subject.append(x)
            x.o = Intersection(p0, None, x, False)
            clip.append(x.o)

            x = Intersection(p1, segment, None, False)
            subject.append(x)
            x.o = Intersection(p1, None, x, True)
            clip.append(x.o)

    if not subject:
        return

    clip.sort(key=cmp_to_key(compare_intersection))
    link(subject)
    link(clip)

    for intersection in clip:
        start_inside = not (start_inside)
        intersection.e = start_inside

    start = subject[0]

    while True:
        current = start
        is_subject = True
        while current.v:
            current = current.n
            if current == start:
                return
        points = current.z
        stream.line_start()

        while True:
            current.v = current.o.v = True
            if current.e:
                if is_subject:
                    for point in points:
                        x, y, _ = point
                        stream.point(x, y)
                else:
                    interpolate(current.x, current.n.x, 1, stream)
                current = current.n
            else:
                if is_subject:
                    points = current.p.z
                    for point in reversed(points):
                        x, y, _ = point
                        stream.point(x, y)
                else:
                    interpolate(current.x, current.p.x, -1, stream)
                current = current.p
            current = current.o
            points = current.z
            is_subject = not is_subject
            if current.v:
                break
        stream.line_end()


def link(array: list[Intersection]):
    n = len(array)
    if n == 0:
        return

    a = array[0]
    for i in range(n):
        b = array[i]
        a.n = b
        b.p = a
        a = b

    b = array[0]
    a.n = b
    b.p = a



from __future__ import annotations

from typing import Any, Generic, overload

from ..types import U, V
from .init import init_range


class ScaleOrdinal(Generic[U, V]):
    """
    Unlike continuous scales, ordinal scales have a discrete domain and range.
    For example, an ordinal scale might map a set of named categories to a set
    of colors, or determine the horizontal positions of columns in a column chart.
    """

    def __init__(self):
        self._index = {}
        self._domain = []
        self._range_vals = []
        self._unknown = None

    def __call__(self, d: U) -> V:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        d : U
            Input value

        Returns
        -------
        V
            Corresponding value from the range
        """
        i = self._index.get(d)
        if i is None:
            if self._unknown is not None:
                return self._unknown
            self._domain.append(d)
            i = len(self._domain) - 1
            self._index[d] = i
        length = len(self._range_vals)
        if not length:
            return None
        index = i % length
        if index >= length or index < 0:
            return None
        return self._range_vals[index]

    def set_domain(self, domain: list[U]) -> ScaleOrdinal:
        """
        Sets the scale’s domain to the specified array of values.

        Parameters
        ----------
        domain : list[U]
            Domain

        Returns
        -------
        ScaleOrdinal
            Itself
        """
        self._domain.clear()
        self._index.clear()
        for value in domain:
            if value in self._index:
                continue
            self._domain.append(value)
            self._index[value] = len(self._domain) - 1
        return self

    def get_domain(self) -> list[U]:
        return self._domain.copy()

    def set_range(self, range_vals: list[V]) -> ScaleOrdinal:
        """
        Sets the scale’s range to the specified array of values

        Parameters
        ----------
        range_vals : list[V]
            Range values

        Returns
        -------
        ScaleOrdinal
            Itself
        """
        self._range_vals = list(range_vals)
        return self

    def get_range(self) -> list[V]:
        return self._range_vals.copy()

    def set_unknown(self, unknown: Any) -> ScaleOrdinal:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        ScaleOrdinal
            Itself
        """
        self._unknown = unknown
        return self

    def get_unknown(self) -> Any:
        return self._unknown

    def copy(self):
        return (
            ScaleOrdinal()
            .set_domain(self.get_domain())
            .set_range(self.get_range())
            .set_unknown(self.get_unknown())
        )

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


@overload
def scale_ordinal() -> ScaleOrdinal[U, V]: ...


@overload
def scale_ordinal(range_vals: list[V]) -> ScaleOrdinal[U, V]: ...


@overload
def scale_ordinal(domain: list[U], range_vals: list[V]) -> ScaleOrdinal[U, V]: ...


def scale_ordinal(*args):
    """
    Build a new ordinal scale with the specified domain and range

    Parameters
    ----------
    domain : list[U]
        Array of values
    range_vals : list[V]
        Array of values

    Returns
    -------
    ScaleOrdinal[U, V]
        Scale object

    Examples
    --------

    >>> scale = d3.scale_ordinal(["a", "b", "c"], ["red", "green", "blue"])
    >>> for c in "abcdefgh":
    ...     print(c, scale(c))
    ...
    ...
    a red
    b green
    c blue
    d red
    e green
    f blue
    g red
    h green
    """
    scale = ScaleOrdinal()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from collections.abc import Callable
from types import MethodType

from .common import PolygonStream


class TransformStream(PolygonStream):
    """
    Default class where methods don't affect the behavior of the passed stream
    object if they are not updated.

    Parameters
    ----------
    stream : PolygonStream
        Stream object
    """

    def __init__(self, stream: PolygonStream):
        self._stream = stream

    def point(self, x: float, y: float):
        """
        Calls the stream :code:`point` method

        Parameters
        ----------
        x : float
            x value
        y : float
            y value
        """
        return self._stream.point(x, y)

    def sphere(self):
        """
        Calls the stream :code:`sphere` method
        """
        return self._stream.sphere()

    def line_start(self):
        """
        Calls the stream :code:`line_start` method
        """
        return self._stream.line_start()

    def line_end(self):
        """
        Calls the stream :code:`line_end` method
        """
        return self._stream.line_end()

    def polygon_start(self):
        """
        Calls the stream :code:`polygon_start` method
        """
        return self._stream.polygon_start()

    def polygon_end(self):
        """
        Calls the stream :code:`polygon_end` method
        """
        return self._stream.polygon_end()

    def __str__(self):
        return f"TransformStream({self._stream})"


class GeoTransformer:
    """
    Transform a stream object given :code:`methods`

    Parameters
    ----------
    methods : dict[str, Callable[..., ...]]
        Methods applied on a stream object
    """

    def __init__(self, methods: dict[str, Callable[..., ...]]):
        self._methods = methods

    def __call__(self, stream: PolygonStream) -> TransformStream:
        """
        Creates a new :code:`TransformStream` with the argument :code:`stream`
        and updates its methods.

        Parameters
        ----------
        stream : PolygonStream
            Stream object

        Returns
        -------
        TransformStream
            Transformed stream
        """
        s = TransformStream(stream)
        for key in self._methods:
            setattr(s, key, MethodType(self._methods[key], s))
        return s

    def __str__(self):
        return f"GeoTransformer({self._methods})"


class GeoTransform:
    """
    Defines an arbitrary transform using the methods defined on the specified
    methods object. Any undefined methods will use pass-through methods that
    propagate inputs to the output stream.

    Parameters
    ----------
    methods : dict[str, Callable[..., ...]]
        Methods applied on a stream object
    """

    def __init__(self, methods: dict[str, Callable[..., ...]]):
        self._methods = methods

    def stream(self) -> GeoTransformer:
        """
        Passes :code:`methods` to a new class which transforms the methods of a
        stream object given the methods and returns it.

        Returns
        -------
        GeoTransformer
            Callable object which modifies the behavior of a stream object
        """
        return GeoTransformer(self._methods)


def geo_transform(methods: dict[str, Callable[..., ...]]) -> GeoTransform:
    """
    Defines an arbitrary transform using the methods defined on the specified
    methods object. Any undefined methods will use pass-through methods that
    propagate inputs to the output stream.

    Parameters
    ----------
    methods : dict[str, Callable[..., ...]]
        Methods applied on a stream object

    Returns
    -------
    GeoTransform
        Class which holds the passed methods
    """
    return GeoTransform(methods)



from __future__ import annotations

from collections.abc import Callable
from math import inf, sqrt

from ..array import argpass
from ..quadtree import quadtree
from ..types import SimulationNode, SimulationNodeFunction
from .constant import constant
from .jiggle import jiggle
from .simulation import x, y


class Apply:
    def __init__(
        self,
        alpha: float,
        strengths: list[float],
        distance_min_2: float,
        distance_max_2: float,
        theta2: float,
        random: Callable[[None], float],
    ):
        self._alpha = alpha
        self._strengths = strengths
        self._distance_min_2 = distance_min_2
        self._distance_max_2 = distance_max_2
        self._theta2 = theta2
        self._random = random
        self._node = None

    def __call__(
        self, quad: list[dict] | dict, x1: float, y1: float, x2: float, y2: float
    ):
        value = quad["value"] if isinstance(quad, dict) else quad[4]["value"]
        if not value:
            return True

        qx = quad["x"] if isinstance(quad, dict) else quad[4]["x"]
        qy = quad["y"] if isinstance(quad, dict) else quad[4]["y"]
        x = qx - self._node["x"]
        y = qy - self._node["y"]
        w = x2 - x1
        length = x * x + y * y

        if w * w / self._theta2 < length:
            if length < self._distance_max_2:
                if x == 0:
                    x = jiggle(self._random)
                    length += x * x
                if y == 0:
                    y = jiggle(self._random)
                    length += y * y
                if length < self._distance_min_2:
                    length = sqrt(self._distance_min_2 * length)
                self._node["vx"] += x * value * self._alpha / length
                self._node["vy"] += y * value * self._alpha / length
            return True
        elif isinstance(quad, list) or length >= self._distance_max_2:
            return
        if quad["data"] != self._node or quad.get("next"):
            if x == 0:
                x = jiggle(self._random)
                length += x * x
            if y == 0:
                y = jiggle(self._random)
                length += y * y
            if length < self._distance_min_2:
                length = sqrt(self._distance_min_2 * length)

        while quad["data"] != self._node:
            w = self._strengths[quad["data"]["index"]] * self._alpha / length
            self._node["vx"] += x * w
            self._node["vy"] += y * w
            quad = quad.get("next")
            if quad is None:
                break

    def update(self, node: SimulationNode):
        self._node = node


class ForceManyBody:
    def __init__(self):
        self._nodes = None
        self._node = None
        self._random = None
        self._alpha = None
        self._strength = constant(-30)
        self._strengths = None
        self._distance_min_2 = 1
        self._distance_max_2 = inf
        self._theta2 = 0.81

    def __call__(self, alpha: float):
        tree = quadtree(self._nodes, x, y).visit_after(self._accumulate)
        self._alpha = alpha
        apply = Apply(
            self._alpha,
            self._strengths,
            self._distance_min_2,
            self._distance_max_2,
            self._theta2,
            self._random,
        )
        for node in self._nodes:
            apply.update(node)
            tree.visit(apply)

    def _initialize(self):
        if self._nodes is None:
            return
        self._strengths = [None] * len(self._nodes)
        for i, node in enumerate(self._nodes):
            self._strengths[node["index"]] = self._strength(node, i, self._nodes)

    def _accumulate(self, quad: list[dict] | dict):
        strength = 0
        weight = 0

        if isinstance(quad, list):
            x = 0
            y = 0
            for q in quad:
                if not q:
                    continue
                if isinstance(q, dict):
                    value = q["value"]
                    qx = q["x"]
                    qy = q["y"]
                else:  # fifth element of a quad as a list
                    value = q[4]["value"]
                    qx = q[4]["x"]
                    qy = q[4]["y"]
                if value:
                    c = abs(value)
                    strength += value
                    weight += c
                    x += c * qx
                    y += c * qy
            if len(quad) == 4:
                quad.append({"x": x / weight, "y": y / weight})
            else:
                quad[4]["x"] = x / weight
                quad[4]["y"] = y / weight
            quad[4]["value"] = strength
        else:
            q = quad
            q["x"] = q["data"]["x"]
            q["y"] = q["data"]["y"]
            while True:
                strength += self._strengths[q["data"]["index"]]
                q = q.get("next")
                if q is None:
                    break
            quad["value"] = strength

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes
        self._random = random
        self._initialize()

    def set_strength(
        self, strength: SimulationNodeFunction[float] | float
    ) -> ForceManyBody:
        """
        Sets the strength accessor to the specified number or function,
        re-evaluates the strength accessor for each node, and returns this
        force. A positive value causes nodes to attract each other, similar to
        gravity, while a negative value causes nodes to repel each other,
        similar to electrostatic charge.

        Parameters
        ----------
        strength : SimulationNodeFunction[float] | float
            Strength function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the strength value (:code:`float`)

        Returns
        -------
        ForceManyBody
            Itself
        """
        if callable(strength):
            self._strength = argpass(strength)
        else:
            self._strength = constant(strength)
        self._initialize()
        return self

    def set_distance_min(self, distance_min: float) -> ForceManyBody:
        """
        Sets the minimum distance between nodes over which this force is
        considered.

        A minimum distance establishes an upper bound on the strength of the
        force between two nearby nodes, avoiding instability. In particular, it
        avoids an infinitely-strong force if two nodes are exactly coincident;
        in this case, the direction of the force is random.

        Parameters
        ----------
        distance_min : float
            Minimum distance value

        Returns
        -------
        ForceManyBody
            Itself
        """
        self._distance_min_2 = distance_min * distance_min
        return self

    def set_distance_max(self, distance_max: float) -> ForceManyBody:
        """
        sets the maximum distance between nodes over which this force is
        considered.

        Specifying a finite maximum distance improves performance and produces
        a more localized layout.

        Parameters
        ----------
        distance_max : float
            Maximum distance value

        Returns
        -------
        ForceManyBody
            Itself
        """
        self._distance_max_2 = distance_max * distance_max
        return self

    def set_theta(self, theta: float) -> ForceManyBody:
        """
        Sets the Barnes-Hut approximation criterion to the specified number and
        returns this force.

        To accelerate computation, this force implements the `Barnes - Hut
        approximation
        <https://en.wikipedia.org/wiki/Barnes%E2%80%93Hut_simulation>`_ which
        takes :math:`O(n \\cdot \\log(n))` per application where :math:`n` is
        the number of nodes. For each application, a quadtree stores the
        current node positions; then for each node, the combined force of all
        other nodes on the given node is computed. For a cluster of nodes that
        is far away, the charge force can be approximated by treating the
        cluster as a single, larger node. The theta parameter determines the
        accuracy of the approximation: if the ratio :math:`w / l` of the width
        w of the quadtree cell to the distance :math:`l` from the node to the
        cell's center of mass is less than :math:`\\theta`, all nodes in the
        given cell are treated as a single node rather than individually.

        Parameters
        ----------
        theta : float
            Barnes-Hut approximation criterion value

        Returns
        -------
        ForceManyBody
            Itself
        """
        self._theta2 = theta * theta
        return self

    def get_strength(self) -> SimulationNodeFunction[float]:
        return self._strength

    def get_distance_min(self) -> float:
        return sqrt(self._distance_min_2)

    def get_distance_max(self) -> float:
        return sqrt(self._distance_max_2)

    def get_theta(self) -> float:
        return sqrt(self._theta2)


def force_many_body() -> ForceManyBody:
    """
    The many-body (or n-body) force applies mutually amongst all nodes. It can
    be used to simulate gravity (attraction) if the strength is positive, or
    electrostatic charge (repulsion) if the strength is negative. This
    implementation uses a quadtree and the Barnes–Hut approximation to greatly
    improve performance; the accuracy can be customized using the theta
    parameter.

    Unlike the link force, which only affect two linked nodes, the charge force
    is global: every node affects every other node, even if they are on
    disconnected subgraphs.

    Returns
    -------
    ForceManyBody
        Force object
    """
    return ForceManyBody()



def format_group(grouping, thousands):
    def group(value, width):
        i = len(value)
        t = []
        j = 0
        g = grouping[0]
        length = 0

        while i > 0 and g > 0:
            if length + g + 1 > width:
                g = max(1, width - length)
            i -= g
            t.append(value[max(0, i) : i + g])
            length += g + 1
            if length > width:
                break
            j = (j + 1) % len(grouping)
            g = grouping[j]

        return thousands.join(reversed(t))

    return group
